from django_webix.forms.forms import WebixForm, WebixModelForm
from django_webix.forms.formsets import WebixInlineFormSet, WebixStackedInlineFormSet, WebixTabularInlineFormSet

__all__ = [
    # forms
    'WebixForm', 'WebixModelForm',
    # formset and inlines
    'WebixInlineFormSet', 'WebixStackedInlineFormSet', 'WebixTabularInlineFormSet'
]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.core.exceptions import ValidationError
from django.forms import BaseInlineFormSet
from django.forms.fields import IntegerField
from django.forms.formsets import TOTAL_FORM_COUNT, INITIAL_FORM_COUNT, MIN_NUM_FORM_COUNT, MAX_NUM_FORM_COUNT
from django.forms.widgets import HiddenInput
from django.utils.functional import cached_property
from django.utils.text import capfirst
from django.utils.translation import ugettext_lazy as _
from extra_views import InlineFormSetFactory

from django_webix.forms import WebixForm, WebixModelForm


class WebixManagementForm(WebixForm):
    """
    ``ManagementForm`` is used to keep track of how many form instances
    are displayed on the page. If adding new forms via javascript, you should
    increment the count field of this form as well.
    """

    def __init__(self, *args, **kwargs):
        self.base_fields[TOTAL_FORM_COUNT] = IntegerField(widget=HiddenInput)
        self.base_fields[INITIAL_FORM_COUNT] = IntegerField(widget=HiddenInput)
        # MIN_NUM_FORM_COUNT and MAX_NUM_FORM_COUNT are output with the rest of
        # the management form, but only for the convenience of client-side
        # code. The POST value of them returned from the client is not checked.
        self.base_fields[MIN_NUM_FORM_COUNT] = IntegerField(required=False, widget=HiddenInput)
        self.base_fields[MAX_NUM_FORM_COUNT] = IntegerField(required=False, widget=HiddenInput)
        super(WebixManagementForm, self).__init__(*args, **kwargs)


class BaseWebixInlineFormSet(BaseInlineFormSet):
    def __init__(self, **kwargs):
        self.request = kwargs.pop('request', None)
        self.container_id = kwargs.pop('container_id', None)
        self.has_add_permission = kwargs.pop('has_add_permission', None)
        self.has_change_permission = kwargs.pop('has_change_permission', None)
        self.has_delete_permission = kwargs.pop('has_delete_permission', None)
        super(BaseWebixInlineFormSet, self).__init__(**kwargs)

    def get_form_kwargs(self, index):
        _form_kwargs = super(BaseWebixInlineFormSet, self).get_form_kwargs(index)
        _form_kwargs.update({
            'request': self.request,
            'inline_id': index,
            'has_add_permission': self.has_add_permission,
            'has_change_permission': self.has_change_permission,
            'has_delete_permission': self.has_delete_permission,
        })
        return _form_kwargs

    def get_rules(self):
        """ Returns a dict with the inline fields rules """

        result = {}
        for form in self.initial_forms:
            result.update(form.get_rules())
        return result

    def get_rules_template(self):
        """Returns the rules template"""

        rules_old = self.form(**{'request':self.request}).get_rules() or {}
        rules_old.pop(self.fk.name, None)  # Remove fk rule
        rules_new = {}
        for key in rules_old:
            rules_new['{}-__prefix__-{}'.format(self.prefix, key)] = rules_old[key]
        return rules_new

    def get_name(self):
        if self.model is not None:
            return capfirst(self.model._meta.verbose_name_plural)
        return ""  # pragma: no cover

    @cached_property
    def management_form(self):
        """Returns the ManagementForm instance for this FormSet."""

        if self.is_bound:
            form = WebixManagementForm(self.data, auto_id=self.auto_id, prefix=self.prefix)
            if not form.is_valid():
                raise ValidationError(
                    _('ManagementForm data is missing or has been tampered with'),
                    code='missing_management_form',
                )
        else:
            form = WebixManagementForm(auto_id=self.auto_id, prefix=self.prefix, initial={
                TOTAL_FORM_COUNT: self.total_form_count(),
                INITIAL_FORM_COUNT: self.initial_form_count(),
                MIN_NUM_FORM_COUNT: self.min_num,
                MAX_NUM_FORM_COUNT: self.max_num
            })
        return form


    def webix_id(self):
        return '{}-group'.format(self.prefix)

    def get_container_id(self):
        return self.container_id

    def get_default_container_id(self):
        return self.webix_id()+'-container'


class WebixInlineFormSet(InlineFormSetFactory):
    template_name = None

    def __init__(self, parent_model, request, instance, view_kwargs=None, view=None, initial=None):
        super(WebixInlineFormSet, self).__init__(parent_model, request, instance, view_kwargs, view)

        # Set initial (for copy purpouse)
        if initial is not None:
            self.initial = initial

        # Set form class
        if self.form_class is None:
            self.form_class = WebixModelForm

        # Set formset class
        if hasattr(self, 'custom_formset_class'):
            self.formset_class = self.custom_formset_class
        else:
            self.formset_class = BaseWebixInlineFormSet
        self.formset_class = type(str('WebixInlineFormSet'), (self.formset_class,), {'template_name': self.template_name})

        # Set queryset
        if hasattr(self, 'get_queryset') and callable(self.get_queryset):
            self.formset_kwargs['queryset'] = self.get_queryset()
        else:
            self.formset_kwargs['queryset'] = self.inline_model.objects.all()

    def get_formset_kwargs(self):
        _formset_kwargs = super(WebixInlineFormSet, self).get_formset_kwargs()
        _formset_kwargs.update({
            'has_add_permission': self.has_add_permission(),
            'has_delete_permission': self.has_delete_permission(),
            'has_change_permission': self.has_change_permission(),
            'request': self.request,
            'container_id': getattr(self, 'container_id', None),
            'initial': self.initial
        })
        return _formset_kwargs

    def get_factory_kwargs(self):
        _factory_kwargs = super(WebixInlineFormSet, self).get_factory_kwargs()
        if self.has_add_permission():
            extra_forms = _factory_kwargs.get('extra', 3)
        else:
            extra_forms = 0
        # FIX for initial data values (list of dict and not instances)
        _factory_kwargs.update({
            'extra': extra_forms #+ len(self.initial),
        })
        return _factory_kwargs

    def has_add_permission(self):
        return True

    def has_change_permission(self):
        return True

    def has_delete_permission(self):
        return True


class WebixStackedInlineFormSet(WebixInlineFormSet):
    template_name = 'django_webix/include/edit_inline/stacked.js'
    style = 'stacked'
    def __init__(self, parent_model, request, instance, view_kwargs=None, view=None, initial=None):
        super(WebixStackedInlineFormSet, self).__init__(parent_model, request, instance, view_kwargs, view, initial)
        self.form_class = type(str('WebixStackedModelForm'), (self.form_class,), {'style': self.style})


class WebixTabularInlineFormSet(WebixInlineFormSet):
    template_name = 'django_webix/include/edit_inline/tabular.js'
    style = 'tabular'
    def __init__(self, parent_model, request, instance, view_kwargs=None, view=None, initial=None):
        super(WebixTabularInlineFormSet, self).__init__(parent_model, request, instance, view_kwargs, view, initial)
        self.form_class = type(str('WebixTabularModelForm'), (self.form_class,), {'style': self.style})



from django.dispatch import Signal

django_webix_view_pre_delete = Signal(providing_args=['instance'])
django_webix_view_post_delete = Signal(providing_args=['instance'])

django_webix_view_pre_save = Signal(providing_args=['instance', 'created', 'form', 'inlines'])
django_webix_view_pre_inline_save = Signal(providing_args=['instance', 'created', 'form', 'inlines'])
django_webix_view_post_save = Signal(providing_args=['instance', 'created', 'form', 'inlines'])



from django.db.models import AutoField
from django.urls import path
from django.utils.text import capfirst

from django_webix.views.generic.base import WebixPermissionsMixin  # utils for permission


class ModelWebixAdmin(WebixPermissionsMixin):
    # WEBIX VIEWS (for fully override)
    create_view = None
    update_view = None
    delete_view = None
    list_view = None

    # JS TEMPLATES
    add_form_template = None
    change_form_template = None
    change_list_template = None
    delete_template = None

    # CREATE AND UPDATE SETTINGS
    enable_button_save_continue = True
    enable_button_save_addanother = True
    enable_button_save_gotolist = True

    # DJANGO WEBIX FORM: OPTION 1
    autocomplete_fields = []
    readonly_fields = []
    fields = None
    exclude = None
    # DJANGO WEBIX FORM: OPTION 2
    form = None

    inlines = []

    # LIST SETTINGS
    ordering = None
    actions = []
    list_display = []

    enable_json_loading = False

    title = None
    actions_style = None
    enable_column_copy = True
    enable_column_delete = True
    enable_row_click = True
    type_row_click = 'single'
    enable_actions = True

    def get_url_pattern_list(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        return '%s.%s.list' % info

    def get_url_pattern_create(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        return '%s.%s.create' % info

    def get_url_pattern_update(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        return '%s.%s.update' % info

    def get_url_pattern_delete(self):
        info = self.model._meta.app_label, self.model._meta.model_name
        return '%s.%s.delete' % info

    def get_model_perms(self, request):
        """
        Return a dict of all perms for this model. This dict has the keys
        ``add``, ``change``, ``delete``, and ``view`` mapping to the True/False
        for each of those actions.
        """
        return {
            'add': self.has_add_permission(request),
            'change': self.has_change_permission(request),
            'delete': self.has_delete_permission(request),
            'view': self.has_view_permission(request),
        }

    def get_field_traverse(self, path):
        _next = self.model
        for j, key in enumerate(path.split('__')):
            if j == 0:
                _next = _next._meta.get_field(key)
            elif hasattr(_next, 'related_model'):
                _next = _next.related_model._meta.get_field(key)
            else:
                raise Exception('TODO?')
        return _next

    def get_list_display(self):
        _fields = []
        for j, field_name in enumerate(self.list_display):
            model_field = self.get_field_traverse(field_name)
            _fields.append({
                'field_name': field_name,
                'datalist_column': '''{{id: "{field_name}", header: ["{header_title}", {{content: "textFilter"}}], {width_adapt}, sort: "string" }}'''.format(
                    field_name=field_name,
                    header_title=capfirst(model_field.verbose_name),
                    width_adapt='fillspace:true' if j == 0 else 'adjust:"all"',
                )
            })
        return _fields

    def __init__(self, model, admin_site):
        self.model = model
        self.opts = model._meta
        self.admin_site = admin_site
        super().__init__()

    def __str__(self):
        return "%s.%s" % (self.model._meta.app_label, self.__class__.__name__)

    def get_queryset(self):
        return self.model._default_manager.all()

    def get_form_fields(self):
        _fields = []
        for _field in self.model._meta.fields:
            if type(_field) != AutoField and \
                (self.exclude is None or _field.name not in self.exclude) and \
                _field.editable is True:
                _fields.append(_field.name)
        return _fields

    def get_form_create_update(self):
        _admin = self
        if self.form:
            return self.form
        else:
            from django_webix.forms import WebixModelForm
            class WebixAdminCreateUpdateForm(WebixModelForm):
                readonly_fields = _admin.readonly_fields
                autocomplete_fields = _admin.autocomplete_fields

                class Meta:
                    localized_fields = ('__all__')
                    model = _admin.model
                    fields = _admin.get_form_fields()

            return WebixAdminCreateUpdateForm

    def get_add_view(self):
        _admin = self
        if self.create_view is not None:
            return self.create_view
        else:
            from django_webix.views import WebixCreateView
            class WebixAdminCreateView(WebixCreateView):
                url_pattern_list = 'admin_webix:' + _admin.get_url_pattern_list()
                url_pattern_create = 'admin_webix:' + _admin.get_url_pattern_create()
                url_pattern_update = 'admin_webix:' + _admin.get_url_pattern_update()
                url_pattern_delete = 'admin_webix:' + _admin.get_url_pattern_delete()

                model = _admin.model
                form_class = _admin.get_form_create_update()
                inlines = _admin.inlines
                model_copy_fields = _admin.get_form_fields()
                enable_button_save_continue = _admin.enable_button_save_continue
                enable_button_save_addanother = _admin.enable_button_save_addanother
                enable_button_save_gotolist = _admin.enable_button_save_gotolist

                def get_queryset(self):
                    return _admin.get_queryset()

            return WebixAdminCreateView

    def get_change_view(self):
        _admin = self
        if self.update_view is not None:
            return self.update_view
        else:
            from django_webix.views import WebixUpdateView
            class WebixAdminUpdateView(WebixUpdateView):
                url_pattern_list = 'admin_webix:' + _admin.get_url_pattern_list()
                url_pattern_create = 'admin_webix:' + _admin.get_url_pattern_create()
                url_pattern_update = 'admin_webix:' + _admin.get_url_pattern_update()
                url_pattern_delete = 'admin_webix:' + _admin.get_url_pattern_delete()

                model = _admin.model
                form_class = _admin.get_form_create_update()
                inlines = _admin.inlines
                model_copy_fields = _admin.get_form_fields()
                enable_button_save_continue = _admin.enable_button_save_continue
                enable_button_save_addanother = _admin.enable_button_save_addanother
                enable_button_save_gotolist = _admin.enable_button_save_gotolist

                def get_queryset(self):
                    return _admin.get_queryset()

            return WebixAdminUpdateView

    def get_delete_view(self):
        _admin = self
        if self.delete_view is not None:
            return self.delete_view
        else:
            from django_webix.views import WebixDeleteView
            class WebixAdminDeleteView(WebixDeleteView):
                url_pattern_list = 'admin_webix:' + _admin.get_url_pattern_list()
                url_pattern_create = 'admin_webix:' + _admin.get_url_pattern_create()
                url_pattern_update = 'admin_webix:' + _admin.get_url_pattern_update()
                url_pattern_delete = 'admin_webix:' + _admin.get_url_pattern_delete()

                model = _admin.model

                def get_queryset(self):
                    return _admin.get_queryset()

            return WebixAdminDeleteView

    def get_list_view(self):
        _admin = self
        if self.list_view is not None:
            return self.list_view
        else:
            from django_webix.views import WebixListView
            class WebixAdminListView(WebixListView):
                url_pattern_list = 'admin_webix:' + _admin.get_url_pattern_list()
                url_pattern_create = 'admin_webix:' + _admin.get_url_pattern_create()
                url_pattern_update = 'admin_webix:' + _admin.get_url_pattern_update()
                url_pattern_delete = 'admin_webix:' + _admin.get_url_pattern_delete()

                model = _admin.model
                order_by = _admin.ordering
                actions = _admin.actions
                enable_json_loading = _admin.enable_json_loading
                title = _admin.title
                actions_style = _admin.actions_style
                enable_column_copy = _admin.enable_column_copy
                enable_column_delete = _admin.enable_column_delete
                enable_row_click = _admin.enable_row_click
                type_row_click = _admin.type_row_click
                enable_actions = _admin.enable_actions

                def get_queryset(self):
                    return super().get_queryset(initial_queryset=_admin.get_queryset())

                def get_fields(self):
                    return _admin.get_list_display()

            return WebixAdminListView

    def get_urls(self):

        # def wrap(view):
        #    def wrapper(*args, **kwargs):
        #        return self.admin_site.admin_view(view)(*args, **kwargs)

        #    wrapper.model_admin = self
        #    return update_wrapper(wrapper, view)

        return [
            path('', self.get_list_view().as_view(), name=self.get_url_pattern_list()),
            path('create/', self.get_add_view().as_view(), name=self.get_url_pattern_create()),
            path('<int:pk>/delete/', self.get_delete_view().as_view(), name=self.get_url_pattern_delete()),
            path('<int:pk>/update/', self.get_change_view().as_view(), name=self.get_url_pattern_update()),
            #            path('', wrap(self.get_list_view), name='%s_%s_changelist' % info),
            #            path('add/', wrap(self.get_add_view), name='%s_%s_add' % info),
            #            path('<path:object_id>/delete/', wrap(self.get_delete_view), name='%s_%s_delete' % info),
            #            path('<path:object_id>/change/', wrap(self.get_change_view), name='%s_%s_change' % info),
        ]

    @property
    def urls(self):
        return self.get_urls()



from django.utils.translation import ugettext as _


def action_config(
    action_key,
    response_type,
    allowed_permissions=[],
    short_description=None,
    modal_title=_("Are you sure you want to proceed with this action?"),
    modal_ok=_("Proceed"),
    modal_cancel=_("Undo"),
):  # TODO: permission check bebore execution

    if response_type not in ['json', 'script', 'blank']:
        raise Exception('No valid response_type [json,script,blank]')

    def decorator(func):
        def wrapper(self, request, qs):
            return func(self, request, qs)

        setattr(wrapper, 'action_key', action_key)
        setattr(wrapper, 'response_type', response_type)
        setattr(wrapper, 'allowed_permissions', allowed_permissions)
        setattr(wrapper, 'short_description', short_description)
        setattr(wrapper, 'modal_title', modal_title)
        setattr(wrapper, 'modal_ok', modal_ok)
        setattr(wrapper, 'modal_cancel', modal_cancel)

        return wrapper

    return decorator



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.exceptions import PermissionDenied
from django.forms import model_to_dict
from django.forms.formsets import all_valid
from django.forms.models import _get_foreign_key
from django.http import HttpResponseRedirect
from django.http import QueryDict
from django.shortcuts import get_object_or_404
from django.utils.encoding import force_text
from django.utils.text import get_text_list
from django.utils.translation import ugettext as _
from extra_views import UpdateWithInlinesView, CreateWithInlinesView

from django_webix.views.generic.base import WebixBaseMixin, WebixPermissionsMixin, WebixUrlMixin
from django_webix.views.generic.signals import (django_webix_view_pre_save,
                                                django_webix_view_pre_inline_save,
                                                django_webix_view_post_save)


class WebixCreateUpdateMixin:
    logs_enable = True

    model = None
    request = None

    enable_button_save_continue = True
    enable_button_save_addanother = True
    enable_button_save_gotolist = True

    template_style = None

    def get_template_style(self):
        _template_style = None
        if self.template_style is None:
            _template_style = 'standard'
        elif self.template_style in ['standard', 'tabs']:
            _template_style = self.template_style
        else:
            raise ImproperlyConfigured(_(
                "Template style is improperly configured"
                " only options are 'standard' or 'tabs' (standard by default)."))
        return _template_style

    def is_enable_button_save_continue(self, request):
        if self.success_url is not None:
            return False
        return self.enable_button_save_continue

    def is_enable_button_save_addanother(self, request):
        if self.success_url is not None:
            return False
        return self.enable_button_save_addanother

    def is_enable_button_save_gotolist(self, request):
        return self.enable_button_save_gotolist

    def get_success_url(self):
        if self.success_url is not None:
            url = self.success_url

        elif self.request.GET.get('_addanother', None) is not None and \
            self.is_enable_button_save_addanother(request=self.request) and \
            self.get_url_create() is not None:
            url = self.get_url_create()

        elif self.request.GET.get('_continue', None) is not None and \
            self.is_enable_button_save_continue(request=self.request) and \
            self.get_url_update(obj=self.object) is not None:
            url = self.get_url_update(obj=self.object)

        elif self.get_url_list() is not None and \
            self.is_enable_button_save_gotolist(request=self.request):  # default
            url = self.get_url_list()

        else:
            raise ImproperlyConfigured(_(
                "No URL to redirect to.  Either provide a url or define"
                " a get_absolute_url method on the Model."))
        return url

    def get_context_data_webix_create_update(self, request, obj=None, **kwargs):

        return {
            # buttons for saving
            'is_enable_button_save_continue': self.is_enable_button_save_continue(request=self.request),
            'is_enable_button_save_addanother': self.is_enable_button_save_addanother(request=self.request),
            'is_enable_button_save_gotolist': self.is_enable_button_save_gotolist(request=self.request),
            # Template style
            'template_style': self.get_template_style(),
        }

    def form_save(self, form):
        self.object = form.save()
        return None

    def inlines_save(self, inlines):
        for formset in inlines:
            formset.save()
        return None

    def forms_valid(self, form, inlines, **kwargs):
        # pre forms valid
        self.pre_forms_valid(form=form, inlines=inlines, **kwargs)
        # validate unique together
        try:
            self.validate_unique_together(form=form, inlines=inlines, **kwargs)
        except ValidationError as e:
            form.add_error(None, str(e))
            return self.forms_invalid(form=form, inlines=inlines, **kwargs)
        # form save
        exception_response = self.form_save(form)
        if exception_response is not None:
            return exception_response
        # post form save
        self.post_form_save(form=form, inlines=inlines, **kwargs)
        # inlines save
        exception_response = self.inlines_save(inlines)
        if exception_response is not None:
            return exception_response
        # post forms valid
        self.post_forms_valid(form=form, inlines=inlines, **kwargs)

        return self.response_valid(success_url=self.get_success_url(), **kwargs)

    def full_validate(self, form, form_validated, inlines, inlines_validated):
        return form, inlines, form_validated and inlines_validated

    def validate_form(self, form):
        if form.is_valid():
            self.object = form.save(commit=False)
            form_validated = True
        else:
            form_validated = False
        return form, form_validated

    def validate_inlines(self, inlines):
        inlines_validated = all_valid(inlines)
        return inlines, inlines_validated

    def response_valid(self, success_url=None, **kwargs):
        return HttpResponseRedirect(success_url)

    def response_invalid(self, form=None, inlines=None, **kwargs):
        return self.render_to_response(self.get_context_data(form=form, inlines=inlines))

    def forms_invalid(self, form, inlines, **kwargs):
        return self.response_invalid(form=form, inlines=inlines, **kwargs)


class WebixCreateView(WebixCreateUpdateMixin,
                      WebixBaseMixin,
                      WebixPermissionsMixin,
                      WebixUrlMixin,
                      CreateWithInlinesView):
    template_name = 'django_webix/generic/create.js'
    model_copy_fields = None
    inlines_copy_fields = None
    send_initial_data = None

    def construct_inlines(self):
        """
        Returns the inline formset instances
        """
        inline_formsets = []
        initial_inlines = self.get_initial_inlines()
        for inline_class in self.get_inlines():
            kwargs = self.kwargs
            initial = initial_inlines.get(inline_class, None)
            inline_instance = inline_class(self.model, self.request, self.object, kwargs, self, initial)
            inline_formset = inline_instance.construct_formset()
            inline_formsets.append(inline_formset)
        return inline_formsets

    def get_form_kwargs(self):
        kwargs = super(WebixCreateView, self).get_form_kwargs()
        kwargs.update({'request': self.request})
        return kwargs

    def get_model_copy_fields(self):
        if self.model_copy_fields is None:
            _model_copy_fields = self.form_class._meta.fields
        else:
            _model_copy_fields = self.model_copy_fields
        return _model_copy_fields

    def get_inlines_copy_fields(self):

        # init
        _inlines_copy_fields = {}
        if self.inlines_copy_fields is None:
            for inline in self.get_inlines():
                _inlines_copy_fields.update({inline: None})
        else:
            _inlines_copy_fields = self.inlines_copy_fields

        full_inlines_copy_fields = {}
        # set fields for each inlines
        for inline, fields in _inlines_copy_fields.items():
            if fields is None:
                fields = []
            full_inlines_copy_fields.update({inline: list(inline.form_class.base_fields.keys())})

        return full_inlines_copy_fields

    def get_initial(self):
        initial = {}
        # main option: send SEND_INITIAL_DATA on (header or GET or POST) and POST initial data
        if self.send_initial_data is not None:
            initial.update(self.send_initial_data)
        # secondary option: send pk_copy for copy instance field values
        elif self.request.GET.get('pk_copy', None) is not None:
            object_to_copy = get_object_or_404(self.get_queryset(), pk=self.request.GET['pk_copy'])
            fields_to_copy = self.get_model_copy_fields()
            if self.model._meta.pk.name in fields_to_copy:
                fields_to_copy.remove(self.model._meta.pk.name)
            initial.update(model_to_dict(object_to_copy,
                                         fields=fields_to_copy))
        return initial

    def get_initial_inlines(self):
        initial_inlines = {}
        if self.request.GET.get('pk_copy', None) is not None:
            object_to_copy = get_object_or_404(self.get_queryset(), pk=self.request.GET['pk_copy'])
            for inline, inline_copy_fields in self.get_inlines_copy_fields().items():
                fields_to_copy = inline_copy_fields
                if inline.model._meta.pk.name in fields_to_copy:
                    fields_to_copy.remove(self.model._meta.pk.name)
                datas = []
                fk = _get_foreign_key(self.model, inline.model)
                for object in inline.model._default_manager.filter(**{fk.name: object_to_copy}):
                    datas.append(model_to_dict(object,
                                               fields=fields_to_copy))
                initial_inlines.update({inline: datas})
        return initial_inlines

    def dispatch(self, *args, **kwargs):
        self.object = None

        # BYPASS POST for initial data
        if self.request.META.get('HTTP_SEND_INITIAL_DATA') or \
           self.request.POST.get('SEND_INITIAL_DATA') or \
           self.request.GET.get('SEND_INITIAL_DATA'):
            if self.request.method == 'POST':
                self.send_initial_data = self.request.POST.dict()
                # switch to GET request
                self.request.method = 'GET'
                self.request.POST = QueryDict('', mutable=True)
            elif self.request.method == 'GET':
                self.send_initial_data = self.request.GET.dict()

        if self.request.method == 'GET':
            if not self.has_view_permission(request=self.request):
                raise PermissionDenied(_('View permission is not allowed'))
        else:
            if not self.has_add_permission(request=self.request):
                raise PermissionDenied(_('Add permission is not allowed'))
        return super(WebixCreateView, self).dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(WebixCreateView, self).get_context_data(**kwargs)
        context.update(self.get_context_data_webix_permissions(request=self.request))
        context.update(self.get_context_data_webix_create_update(request=self.request))
        context.update(self.get_context_data_webix_url(request=self.request))
        context.update(self.get_context_data_webix_base(request=self.request))
        return context

    def pre_forms_valid(self, form=None, inlines=None, **kwargs):
        '''
        Before all data saving
        '''
        django_webix_view_pre_save.send(sender=self,
                                        instance=None,
                                        created=True,
                                        form=form,
                                        inlines=inlines)

    def post_form_save(self, form=None, inlines=None, **kwargs):
        '''
        After form save and before inlines save
        '''
        django_webix_view_pre_inline_save.send(sender=self,
                                               instance=self.object,
                                               created=True,
                                               form=form,
                                               inlines=inlines)

    def post_forms_valid(self, form=None, inlines=None, **kwargs):
        '''
        After all data saved
        '''
        django_webix_view_post_save.send(sender=self,
                                         instance=self.object,
                                         created=True,
                                         form=form,
                                         inlines=inlines)
        # LOG
        anonymous = self.request.user.is_anonymous() if callable(
            self.request.user.is_anonymous) else self.request.user.is_anonymous
        if self.logs_enable is True and not anonymous and apps.is_installed('django.contrib.admin'):
            from django.contrib.admin.models import LogEntry, ADDITION
            from django.contrib.contenttypes.models import ContentType
            LogEntry.objects.log_action(
                user_id=self.request.user.pk,
                content_type_id=ContentType.objects.get_for_model(self.object).pk,
                object_id=self.object.pk,
                object_repr=force_text(self.object),
                action_flag=ADDITION
            )


    def validate_unique_together(self, form=None, inlines=None, **kwargs):
        # self.object.validate_unique()
        if form is not None:
            form.instance.validate_unique()

class WebixUpdateView(WebixCreateUpdateMixin, WebixBaseMixin, WebixPermissionsMixin, WebixUrlMixin,
                      UpdateWithInlinesView):
    template_name = 'django_webix/generic/update.js'

    def get_form_kwargs(self):
        kwargs = super(WebixUpdateView, self).get_form_kwargs()
        kwargs.update({'request': self.request})
        return kwargs

    def get_object(self, queryset=None):
        if getattr(self, 'object', None) is not None:
            return self.object
        return super(WebixUpdateView, self).get_object(queryset=queryset)

    def dispatch(self, *args, **kwargs):
        self.object = self.get_object()
        if self.request.method == 'GET':
            if not self.has_view_permission(request=self.request, obj=self.object):
                raise PermissionDenied(_('View permission is not allowed'))
        else:
            if not self.has_change_permission(request=self.request, obj=self.object):
                raise PermissionDenied(_('Change permission is not allowed'))
        return super(WebixUpdateView, self).dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(WebixUpdateView, self).get_context_data(**kwargs)
        context.update(self.get_context_data_webix_permissions(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_create_update(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_url(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_base(request=self.request))
        return context

    def validate_unique_together(self, form=None, inlines=None, **kwargs):
        self.object.validate_unique()
        if form is not None:
            form.instance.validate_unique()

    def pre_forms_valid(self, form=None, inlines=None, **kwargs):
        '''
        Before all data saving
        '''
        django_webix_view_pre_save.send(sender=self,
                                        instance=self.object,
                                        created=False,
                                        form=form,
                                        inlines=inlines)

    def post_form_save(self, form=None, inlines=None, **kwargs):
        '''
        After form save and before inlines save
        '''

        django_webix_view_pre_inline_save.send(sender=self,
                                               instance=self.object,
                                               created=False,
                                               form=form,
                                               inlines=inlines)

    def post_forms_valid(self, form=None, inlines=None, **kwargs):

        django_webix_view_post_save.send(sender=self,
                                         instance=self.object,
                                         created=False,
                                         form=form,
                                         inlines=inlines)

        anonymous = self.request.user.is_anonymous() if callable(
            self.request.user.is_anonymous) else self.request.user.is_anonymous
        if self.logs_enable is True and not anonymous and apps.is_installed('django.contrib.admin'):
            from django.contrib.admin.models import LogEntry, CHANGE
            from django.contrib.contenttypes.models import ContentType
            LogEntry.objects.log_action(
                user_id=self.request.user.pk,
                content_type_id=ContentType.objects.get_for_model(self.object).pk,
                object_id=self.object.pk,
                object_repr=force_text(self.object),
                action_flag=CHANGE,
                change_message=_('Changed %s.') % get_text_list(form.changed_data, _('and'))
            )


class WebixCreateWithInlinesView(WebixCreateView):
    def __init__(self, *args, **kwargs):
        from warnings import warn
        warn('`django_webix.views.WebixCreateWithInlinesView` has been renamed to `WebixCreateView`. '
             '`WebixCreateWithInlinesView` will be removed in a future release.', DeprecationWarning)
        super(WebixCreateWithInlinesView, self).__init__(*args, **kwargs)


class WebixCreateWithInlinesUnmergedView(WebixCreateView):
    style = 'unmerged'

    def __init__(self, *args, **kwargs):
        from warnings import warn
        warn('`django_webix.views.WebixCreateWithInlinesUnmergedView` has been renamed to `WebixCreateView`. '
             '`WebixCreateWithInlinesUnmergedView` will be removed in a future release.', DeprecationWarning)
        super(WebixCreateWithInlinesUnmergedView, self).__init__(*args, **kwargs)


class WebixUpdateWithInlinesView(WebixUpdateView):
    def __init__(self, *args, **kwargs):
        from warnings import warn
        warn('`django_webix.views.WebixUpdateWithInlinesView` has been renamed to `WebixUpdateView`. '
             '`WebixUpdateWithInlinesView` will be removed in a future release.', DeprecationWarning)
        super(WebixUpdateWithInlinesView, self).__init__(*args, **kwargs)


class WebixUpdateWithInlinesUnmergedView(WebixUpdateView):
    style = 'unmerged'

    def __init__(self, *args, **kwargs):
        from warnings import warn
        warn('`django_webix.views.WebixUpdateWithInlinesUnmergedView` has been renamed to `WebixUpdateView`. '
             '`WebixUpdateWithInlinesUnmergedView` will be removed in a future release.', DeprecationWarning)
        super(WebixUpdateWithInlinesUnmergedView, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import re
import six
from django import template
from django.conf import settings
import datetime

register = template.Library()


@register.filter('get_value_from_dict')
def get_value_from_dict(dict_data, key):
    """
    usage example {{ your_dict|get_value_from_dict:your_key }}
    """
    if key:
        return dict_data.get(key)


@register.simple_tag(name='webix_version')
def webix_version():
    return settings.WEBIX_VERSION


@register.simple_tag(name='webix_license')
def webix_license():
    return settings.WEBIX_LICENSE


@register.simple_tag(name='webix_fontawesome_css_url')
def webix_fontawesome_css_url():
    if hasattr(settings, 'WEBIX_FONTAWESOME_CSS_URL'):
        return settings.WEBIX_FONTAWESOME_CSS_URL
    else:
        return 'django_webix/fontawesome-5.7.2/css/all.min.css'

@register.simple_tag(name='webix_fontawesome_version')
def webix_fontawesome_version():
    if hasattr(settings, 'WEBIX_FONTAWESOME_VERSION'):
        return settings.WEBIX_FONTAWESOME_VERSION
    else:
        return '5.7.2'


@register.filter
def comma_to_underscore(value):
    return value.replace(".", "_")


@register.filter_function
def order_by(queryset, args):
    args = [x.strip() for x in args.split(',')]
    return queryset.order_by(*args)


@register.filter
def format_list_value(value):
    if type(value)==datetime.date:
        return value.strftime('%d/%m/%Y')
    elif type(value) == datetime.datetime:
        return value.strftime('%d/%m/%Y %H:%M')
    elif value is None:
        return ''
    return str(value)


@register.filter
def getattr(obj, args):
    """ Try to get an attribute from an object.

    Example: {% if block|getattr:"editable,True" %}

    Beware that the default is always a string, if you want this
    to return False, pass an empty second argument:
    {% if block|getattr:"editable," %}
    """
    splitargs = args.split(',')
    try:
        (attribute, default) = splitargs
    except ValueError:
        (attribute, default) = args, ''

    if isinstance(obj, six.string_types):
        return obj

    try:
        attr = obj.__getattribute__(attribute)
    except AttributeError:
        attr = obj.__dict__.get(attribute, default)
    except Exception:
        attr = default

    if hasattr(attr, '__call__'):
        return attr.__call__()
    else:
        return attr


class SetVarNode(template.Node):
    def __init__(self, new_val, var_name):
        self.new_val = new_val
        self.var_name = var_name

    def render(self, context):
        context[self.var_name] = self.new_val
        return ''


@register.tag
def setvar(parser, token):
    # This version uses a regular expression to parse tag contents.
    try:
        # Splitting by None == splitting by spaces.
        tag_name, arg = token.contents.split(None, 1)
    except ValueError:
        raise template.TemplateSyntaxError("%r tag requires arguments" % token.contents.split()[0])
    m = re.search(r'(.*?) as (\w+)', arg)
    if not m:
        raise template.TemplateSyntaxError("%r tag had invalid arguments" % tag_name)
    new_val, var_name = m.groups()
    if not (new_val[0] == new_val[-1] and new_val[0] in ('"', "'")):
        raise template.TemplateSyntaxError("%r tag's argument should be in quotes" % tag_name)
    return SetVarNode(new_val[1:-1], var_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import json

from django import forms
from django.views.generic import TemplateView, FormView

from django_webix.forms import WebixModelForm, WebixForm
from django_webix.forms import WebixTabularInlineFormSet, WebixStackedInlineFormSet, BaseWebixInlineFormSet
from django_webix.views import WebixCreateWithInlinesView, WebixCreateWithInlinesUnmergedView, \
    WebixUpdateWithInlinesView, WebixUpdateWithInlinesUnmergedView, \
    WebixDeleteView
from .models import MyModel, InlineModel, InlineStackedModel, InlineEmptyModel, UrlsModel


class MyLoginForm(WebixForm):
    username = forms.CharField(required=True)
    password = forms.CharField(widget=forms.PasswordInput(), required=True)


class MyLoginView(FormView):
    form_class = MyLoginForm
    template_name = 'django_webix/generic/create.js'


class InlineModelInline(WebixTabularInlineFormSet):
    model = InlineModel
    fields = '__all__'

    def get_queryset(self):
        return self.inline_model.objects.all()


class InlineStackedModelForm(WebixModelForm):
    autocomplete_fields = ['inlinemodel']
    min_count_suggest = 0

    class Meta:
        model = InlineStackedModel
        fields = '__all__'


class InlineStackedModelInline(WebixStackedInlineFormSet):
    model = InlineStackedModel
    fields = '__all__'
    form_class = InlineStackedModelForm


class InlineEmptyModelForm(WebixModelForm):
    factory_kwargs = {
        'extra': 0
    }


class InlineEmptyModelFormSet(BaseWebixInlineFormSet):
    def get_queryset(self):
        return self.queryset


class InlineEmptyModelInline(WebixTabularInlineFormSet):
    model = InlineEmptyModel
    fields = '__all__'
    form_class = InlineEmptyModelForm
    custom_formset_class = InlineEmptyModelFormSet


class MyModelForm(WebixModelForm):
    readonly_fields = ['readonly', 'raise_error', 'datetimefield', 'datefield', 'booleanfield', 'datefield_empty',
                       'datetimefield_empty']

    class Meta:
        model = MyModel
        fields = '__all__'


class MyModelListView(TemplateView):
    template_name = 'list.js'

    def get_context_data(self, **kwargs):
        context = super(MyModelListView, self).get_context_data(**kwargs)
        context['datalist'] = json.dumps([{
            'id': i.pk,
            'field': i.field
        } for i in MyModel.objects.all()])
        return context


class MyModelCreateView(WebixCreateWithInlinesView):
    model = MyModel
    inlines = [InlineModelInline, InlineStackedModelInline, InlineEmptyModelInline]
    form_class = MyModelForm
    permissions = False


class MyModelCreateUnmergedView(WebixCreateWithInlinesUnmergedView):
    model = MyModel
    inlines = [InlineModelInline, InlineStackedModelInline, InlineEmptyModelInline]
    form_class = MyModelForm
    permissions = False

    url_create = 'app_name.mymodel.create_unmerged'


class MyModelCreateErrorView(WebixCreateWithInlinesView):
    model = MyModel
    inlines = [InlineModelInline, InlineStackedModelInline, InlineEmptyModelInline]
    form_class = MyModelForm
    style = 'error'

    url_create = 'app_name.mymodel.create_error'


class MyModelUpdateView(WebixUpdateWithInlinesView):
    model = MyModel
    inlines = [InlineModelInline, InlineStackedModelInline, InlineEmptyModelInline]
    form_class = MyModelForm


class MyModelUpdateErrorView(WebixUpdateWithInlinesView):
    model = MyModel
    inlines = [InlineModelInline, InlineStackedModelInline, InlineEmptyModelInline]
    form_class = MyModelForm
    style = 'error'

    url_create = 'app_name.mymodel.update_error'


class MyModelDeleteView(WebixDeleteView):
    model = MyModel


class InlineModelUpdateView(WebixUpdateWithInlinesUnmergedView):
    model = InlineModel
    fields = '__all__'


class InlineStackedModelDelete(WebixDeleteView):
    model = InlineStackedModel


class CreateSuccessUrlView(WebixCreateWithInlinesView):
    model = UrlsModel
    fields = '__all__'
    success_url = '/'


class CreateUrlUpdateView(WebixCreateWithInlinesView):
    model = UrlsModel
    fields = '__all__'
    url_update = 'app_name.mymodel.create_urlupdate'


class CreateUrlListView(WebixCreateWithInlinesView):
    model = UrlsModel
    fields = '__all__'
    url_update = None
    url_list = 'app_name.mymodel.create_urllist'


class CreateNoUrlView(WebixCreateWithInlinesView):
    model = UrlsModel
    fields = '__all__'
    url_update = None
    url_list = None


class UpdateSuccessUrlView(WebixUpdateWithInlinesView):
    model = UrlsModel
    fields = '__all__'
    success_url = '/'


class UpdateUrlUpdateView(WebixUpdateWithInlinesView):
    model = UrlsModel
    fields = '__all__'
    url_update = 'app_name.mymodel.update_urlupdate'


class UpdateNoUrlView(WebixUpdateWithInlinesView):
    model = UrlsModel
    fields = '__all__'
    url_update = None
    url_list = None


class DeleteSuccessUrlView(WebixDeleteView):
    model = UrlsModel
    success_url = '/'


class DeleteUrlListView(WebixDeleteView):
    model = UrlsModel
    url_list = 'app_name.mymodel.delete_urllist'


class DeleteNoUrlView(WebixDeleteView):
    model = UrlsModel
    url_update = None
    url_list = None



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import json
import operator
from functools import reduce

from django.apps import apps
from django.conf import settings
from django.contrib.admin.utils import prepare_lookup_value
from django.core.exceptions import PermissionDenied
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, connection
from django.db.models import Q, QuerySet
from django.db.models.constants import LOOKUP_SEP
from django.http import HttpResponse
from django.utils.encoding import smart_text
from django.views import View
from django.views.decorators.cache import never_cache

AUTOCOMPLETE_SEARCH_FIELDS = getattr(settings, "WEBIX_AUTOCOMPLETE_SEARCH_FIELDS", {})


def get_label(f):
    return f.related_label() if getattr(f, "related_label", None) else smart_text(f)


def ajax_response(data):
    return HttpResponse(json.dumps(data, cls=DjangoJSONEncoder), content_type='application/javascript')


class RelatedLookup(View):
    "Related Lookup"

    def check_user_permission(self):
        """ CUSTOM """

        if not (self.request.user.is_active and self.request.user.is_authenticated):
            raise PermissionDenied

    def get_model(self):
        try:
            self.model = apps.get_model(self.GET['app_label'], self.GET['model_name'])
        except LookupError:
            self.model = None
        return self.model

    def get_filtered_queryset(self, qs):
        """ CUSTOM """

        query_string = self.GET.get('query_string', None)
        qset = Q()
        if query_string:
            for j, item2 in enumerate(query_string.split("|")):  # OR
                filters = {}
                for item in item2.split(":"):  # AND
                    k, v = item.split("=")
                    VALUE = smart_text(v)
                    if VALUE == 'TRUE':
                        VALUE = True
                    elif VALUE == 'FALSE':
                        VALUE = False
                    elif 'None' in VALUE:
                        VALUE = False
                    if k != "_to_field":
                        filters[smart_text(k)] = prepare_lookup_value(smart_text(k), VALUE)
                qset |= Q(**filters)
        try:
            qs = qs.filter(qset)
        except:
            qs = qs.none()
        return qs

    def get_queryset(self):
        qs = self.model._default_manager.get_queryset()
        # model_admin = self.get_model_admin()
        # if model_admin is not None:
        #     qs = model_admin.get_queryset(self.request)
        qs = self.get_filtered_queryset(qs)
        return qs


class AutocompleteWebixLookup(RelatedLookup):
    "AutocompleteLookup"

    def request_is_valid(self):
        return 'filter[value]' in self.GET and 'app_label' in self.GET and 'model_name' in self.GET

    def get_searched_queryset(self, qs):
        """ CUSTOM """

        model = self.model
        term = self.GET["filter[value]"]  # ho cambiato da "term" a "filter[value]"
        try:
            term = model.autocomplete_term_adjust(term)
        except AttributeError:
            pass

        try:
            search_fields = model.autocomplete_search_fields()
        except AttributeError:
            try:
                search_fields = AUTOCOMPLETE_SEARCH_FIELDS[model._meta.app_label][model._meta.module_name]
            except KeyError:
                search_fields = ()

        if search_fields:
            for word in term.split():
                search = [models.Q(**{smart_text(item): smart_text(word)}) for item in search_fields]
                search_qs = QuerySet(model)
                search_qs.query.select_related = qs.query.select_related
                search_qs = search_qs.filter(reduce(operator.or_, search))
                qs &= search_qs
        else:
            qs = model.objects.none()
        return qs

    def get_final_ordering(self, model, previous_lookup_parts=None):
        """
        This recursive function returns the final lookups
        for the default ordering of a model.

        Considering the models below, `get_final_ordering(Book)` will return
        `['-type__name', 'name']` instead of the simple `['-type', 'name']`
        one would get using `Book._meta.ordering`.

            class BookType(Model):
                name = CharField(max_length=50)

                class Meta:
                    ordering = ['name']

            class Book(Model):
                name = CharField(max_length=50)
                type = ForeignKey(BookType)

                class Meta:
                    ordering = ['-type', 'name']
        """
        ordering = []
        for lookup in model._meta.ordering:
            opts = model._meta
            for part in lookup.lstrip('-').split(LOOKUP_SEP):
                field = opts.get_field(part)
                if field.is_relation:
                    opts = field.rel.to._meta if hasattr(field, 'rel') else field.related_model._meta
            if previous_lookup_parts is not None:
                lookup = previous_lookup_parts + LOOKUP_SEP + lookup
            if field.is_relation:
                ordering.extend(self.get_final_ordering(opts.model, lookup))
            else:
                ordering.append(lookup)
        return ordering

    def get_queryset(self):
        qs = super(AutocompleteWebixLookup, self).get_queryset()
        qs = self.get_filtered_queryset(qs)
        qs = self.get_searched_queryset(qs)

        if connection.vendor == 'postgresql':
            ordering = self.get_final_ordering(self.model)
            distinct_columns = [o.lstrip('-') for o in ordering]
            pk_name = self.model._meta.pk.name
            if pk_name not in distinct_columns:
                distinct_columns.append(pk_name)
            return qs.order_by(*ordering).distinct(*distinct_columns)

        return qs.distinct()

    def get_data(self):
        to_field = self.GET.get("to_field", 'pk')
        if self.GET.get("nolimit", None):
            _AUTOCOMPLETE_LIMIT = 1000
        else:
            _AUTOCOMPLETE_LIMIT = 50
        return [{
            "id": '%s' % getattr(f, to_field),
            "value": get_label(f)
        } for f in self.get_queryset()[:_AUTOCOMPLETE_LIMIT]]

    @never_cache
    def get(self, request, *args, **kwargs):
        self.check_user_permission()
        self.GET = self.request.GET

        if self.request_is_valid():
            self.get_model()
            if self.model is not None:
                data = self.get_data()
                if data:
                    return ajax_response(data)

        # overcomplicated label translation
        data = []
        return ajax_response(data)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import copy
from collections import OrderedDict, defaultdict
from json import dumps, loads

import django
import six
from django import forms
from django.conf import settings

try:
    from django.contrib.gis.geos import GEOSGeometry
except ImportError:
    GEOSGeometry = object

from django.core.exceptions import FieldDoesNotExist
from django.core.exceptions import ImproperlyConfigured
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models, connection
from django.forms.forms import DeclarativeFieldsMetaclass
from django.forms.models import ModelFormMetaclass, InlineForeignKeyField
from django.forms.utils import ErrorList
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.text import capfirst
from django.utils.translation import ugettext_lazy as _
from random import randint
from sorl.thumbnail import get_thumbnail

try:
    from django.contrib.postgres.forms.jsonb import JSONField
except ImportError:
    JSONField = forms.Field

try:
    from django.contrib.postgres.fields.array import SimpleArrayField
except ImportError:
    SimpleArrayField = None

try:
    from django.contrib.gis.forms import GeometryField
except ImportError:
    GeometryField = None


class BaseWebixMixin(object):
    form_fix_height = None
    min_count_suggest = 100
    style = 'stacked'
    readonly_fields = []
    autocomplete_fields = []
    autocomplete_fields_exclude = []
    autocomplete_fields_urls = {}

    class Meta:
        localized_fields = '__all__'  # to use comma as separator in i18n

    def get_fields_data_with_prefix(self, data=None):
        readonly_fields = self.get_readonly_fields()
        if data is not None:
            qdict = data.copy()
            for name, field in copy.deepcopy(self.base_fields).items():
                key = self.add_prefix(name)
                if key in data and \
                    name not in readonly_fields:
                    if isinstance(field, forms.models.ModelMultipleChoiceField):
                        temp_list = []
                        if type(data[key]) != list:
                            for val in data[key].split(','):
                                if val != '':
                                    temp_list.append(val)
                        else:
                            temp_list = data[key]
                        qdict.setlist(key, temp_list)
                    else:
                        val = data[key]
                        qdict.update({key: val if val not in ['null', u'null'] else None})
            data = qdict
        return data

    def set_readonly_fields(self):
        # Replace field if it is in readonly_fields list
        for readonly_field in self.get_readonly_fields():
            try:
                _field = self._meta.model._meta.get_field(readonly_field)
            except FieldDoesNotExist:
                _field = None
            else:
                del self.fields[readonly_field]
                if not hasattr(self, 'instance') or self.instance.pk is None:
                    self.fields[readonly_field] = forms.CharField(
                        label=_(_field.verbose_name).capitalize(), required=False
                    )
                else:
                    if isinstance(_field, django.db.models.fields.DateTimeField) or \
                        isinstance(_field, django.forms.fields.DateTimeField) or \
                        isinstance(_field, django.forms.fields.DateTimeInput):
                        # TODO: it can be improved
                        self.fields[readonly_field] = forms.CharField(
                            label=_(_field.verbose_name).capitalize(),
                            required=False
                        )
                        if getattr(self.instance, readonly_field):
                            value = '{}'.format(getattr(self.instance, readonly_field).strftime('%d/%m/%Y %H:%i'))
                        else:
                            value = ''
                    elif isinstance(_field, django.db.models.fields.DateField) or \
                        isinstance(_field, django.forms.fields.DateField):
                        # TODO: it can be improved
                        self.fields[readonly_field] = forms.CharField(
                            label=_(_field.verbose_name).capitalize(),
                            required=False
                        )
                        if getattr(self.instance, readonly_field):
                            value = '{}'.format(getattr(self.instance, readonly_field).strftime('%d/%m/%Y'))
                        else:
                            value = ''
                    elif isinstance(_field, django.db.models.fields.BooleanField) or \
                        isinstance(_field, django.forms.fields.CheckboxInput):
                        # TODO: it can be improved
                        self.fields[readonly_field] = forms.BooleanField(
                            label=_(_field.verbose_name).capitalize(),
                            required=False
                        )
                        value = getattr(self.instance, readonly_field) or None
                    else:
                        self.fields[readonly_field] = forms.CharField(
                            label=_(_field.verbose_name).capitalize(),
                            required=False
                        )
                        value = '{}'.format(getattr(self.instance, readonly_field) or '')
                    self.fields[readonly_field].initial = value
                    self.initial[readonly_field] = value

    def webix_extra_clean(self, cleaned_data):
        for key, value in cleaned_data.items():
            if isinstance(value, six.string_types):
                cleaned_data[key] = value.strip()
        for _readonly_field in self.get_readonly_fields():
            cleaned_data.pop(_readonly_field, None)
        return cleaned_data

    def get_readonly_fields(self):
        """ Returns a list of readonly fields """
        if self.has_change_permission == False:
            readonly_fields = []
            for field_name in self.base_fields.keys():
                if not field_name.endswith('_block'):
                    readonly_fields.append(field_name)
            return readonly_fields
        else:
            return self.readonly_fields

    def _add_null_choice(self, choices):
        return [option for option in choices if option['id'] not in ['', u'', None]]

    def _get_url_suggest(self, app_label, model_name, to_field_name=None):
        """ Returns the url to autocomplete model choiche field """

        url = "{url}?app_label={app_label}&model_name={model_name}"
        if to_field_name:
            url += "&to_field={to_field}"
        return url.format(
            url=reverse('webix_autocomplete_lookup'),
            model_name=model_name,
            app_label=app_label,
            to_field=to_field_name,
        )

    @property
    def webix_id(self):
        """ Returns the form id """

        if hasattr(self.Meta, 'model'):
            return '{app_label}.{model_name}'.format(
                app_label=self.Meta.model._meta.app_label,
                model_name=self.Meta.model._meta.model_name
            )
        return self.__class__.__name__

    def get_rules(self):
        """ Returns the form rules """

        rules = defaultdict(list)
        for name, field in self.fields.items():
            if field.required and (not type(field) == forms.FileField or field.initial is not None):
                rules[self.add_prefix(name)].append({'rule': 'isNotEmpty'})
            if isinstance(field, forms.EmailField):
                rules[self.add_prefix(name)].append({'rule': 'isEmail'})
            elif isinstance(field, forms.FloatField):
                rules[self.add_prefix(name)].append({'rule': 'isNumber', 'min': -six.MAXSIZE, 'max': six.MAXSIZE})
            elif isinstance(field, forms.DecimalField):
                rules[self.add_prefix(name)].append({'rule': 'isNumber', 'min': -six.MAXSIZE, 'max': six.MAXSIZE})
            elif isinstance(field, forms.IntegerField):
                rules[self.add_prefix(name)].append({'rule': 'isInteger', 'min': -six.MAXSIZE, 'max': six.MAXSIZE})
            elif isinstance(field, forms.CharField):
                rules[self.add_prefix(name)].append({'rule': 'isString', 'min': 0, 'max': field.max_length})
        return dict(rules)

    @property
    def get_elements(self):
        """ Return an OrderedDict with all form fields in Webix format """

        elements = OrderedDict()
        for name, field in self.fields.items():
            _pass = False
            label = force_text(field.label).capitalize()
            if field.required:
                label = '<strong>{}</strong>'.format(label)
            el = {
                'view': 'text',
                'label': label,
                'name': self.add_prefix(name),
                'id': self[name].auto_id,
                'labelWidth': 200,
                'django_type_field': str(type(field).__name__),
            }

            if name in self.get_readonly_fields():
                el.update({
                    'disabled': True,
                    # 'id': '{}.{}'.format(self.webix_id, self[name].auto_id) # probabily old: disabled for has_change_permission=False
                })

            # Get initial value
            if hasattr(self, 'cleaned_data'):
                if type(field) == forms.models.ModelMultipleChoiceField:
                    initial = self.data.getlist(self.add_prefix(name), field.initial)
                elif connection.vendor == 'postgresql' and isinstance(field, JSONField):
                    initial = self.data.get(self.add_prefix(name), None)
                    if initial is not None:
                        initial = loads(initial)
                    else:
                        initial = field.initial
                else:
                    initial = self.data.get(self.add_prefix(name), field.initial)
            else:
                initial = self.initial.get(name, field.initial)

            # EmailField
            if isinstance(field, forms.EmailField):
                el.update({
                    'view': 'text'
                })
                if initial is not None:
                    el.update({'value': initial})
            # FloatField
            elif isinstance(field, forms.FloatField):
                if initial is not None:
                    el.update({'value': '{}'.format(initial).replace('.', ',')})
            # DecimalField
            elif isinstance(field, forms.DecimalField):
                # el.update({'view':''})
                if initial is not None:
                    el.update({'value': '{}'.format(initial).replace('.', ',')})
            # IntegerField
            elif isinstance(field, forms.IntegerField):
                el.update({
                    'view': 'text'
                })
                if initial is not None and initial not in ['', None]:
                    el.update({'value': '{}'.format(initial)})
            # TimeField
            elif isinstance(field, forms.TimeField):
                el.update({
                    'view': "datepicker",
                    'type': "time",
                    'stringResult': True,
                    'editable': True
                })
                if initial is not None:
                    if isinstance(initial, six.string_types):
                        el.update({
                            'value': '{}'.format(initial).replace('-', ',').replace(' ', ',').replace(':', ',')
                        })
                    elif callable(initial):
                        el.update({'value': '{}'.format(initial().strftime('%H,%M'))})
                    else:
                        el.update({'value': '{}'.format(initial.strftime('%H,%M'))})
            # DateField
            elif isinstance(field, forms.DateField):
                el.update({
                    'view': 'datepicker',
                    'format': "%d/%m/%Y",
                    'stringResult': True,
                    'editable': True
                })
                if initial is not None:
                    if isinstance(initial, six.string_types):
                        el.update({'value': '{}'.format(initial[0:10]).replace('-', ',')})
                    elif callable(initial):
                        el.update({'value': '{}'.format(initial().strftime('%Y,%m,%d'))})
                    else:
                        el.update({'value': '{}'.format(initial.strftime('%Y,%m,%d'))})
            # DateTimeField
            elif isinstance(field, forms.DateTimeField):
                el.update({
                    'view': "datepicker",
                    'format': "%d/%m/%Y %H:%i",
                    'stringResult': True,
                    'timepicker': True,
                    'editable': True
                })
                if initial is not None:
                    if isinstance(initial, six.string_types):
                        el.update({
                            'value': '{}'.format(initial).replace('-', ',').replace(' ', ',').replace(':', ',')
                        })
                    elif callable(initial):
                        el.update({'value': '{}'.format(initial().strftime('%Y,%m,%d,%H,%M'))})
                    else:
                        el.update({'value': '{}'.format(initial.strftime('%Y,%m,%d,%H,%M'))})
            # BooleanField NullBooleanField
            elif isinstance(field, forms.NullBooleanField) or isinstance(field, forms.BooleanField):
                el.update({
                    'view': 'checkbox',
                    'checkValue': '2',
                    'uncheckValue': ''
                })
                if initial is not None:
                    if (isinstance(initial, bool) and initial) or \
                        (isinstance(initial, six.string_types) and initial.lower() == 'true') or \
                        (isinstance(initial, int) and initial == 2):
                        el.update({'value': '2'})
            # URLField
            elif isinstance(field, forms.URLField):
                # el.update({'view':''})
                if initial is not None:
                    el.update({'value': initial})
            # SlugField
            elif isinstance(field, forms.SlugField):
                # el.update({'view':''})
                if initial is not None:
                    el.update({'value': initial})
            # Image  # TODO
            elif 'image' in str(type(field)).lower():
                if initial is not None:
                    el.update({'value': None})
                el.update({
                    'view': 'uploader',
                    'autosend': False,
                    'multiple': False,
                    'width': 100,
                    'label': _('Upload new image')
                })
                if initial:
                    img_small = get_thumbnail(initial, '150x100', quality=85)
                    img_big = get_thumbnail(initial, '500x400', quality=85)
                    key = randint(1, 100000)
                    _template_file = '<img src="{}"  onclick="image_modal(\'{}\',{},{},\'{}\')">'.format(
                        img_small.url, img_big.url, 500, 400, str(key)
                    )
                else:
                    _template_file = ''
                elements.update({
                    '{}_block'.format(self[name].html_name): {
                        'cols': [
                            {
                                'name_label': name,
                                'id_label': 'label_' + name,
                                'borderless': True,
                                'template': label,
                                'height': 30,
                                'width': 200,
                                'css': {'background-color': 'transparent !important'}
                            },
                            {
                                'name_label': name,
                                'id_label': 'preview_' + name,
                                'borderless': True,
                                'template': _template_file,
                                'height': 100,
                                'width': 170
                            },
                            el,
                            {
                                'borderless': True,
                                'template': '',
                                'height': 30
                            },
                        ]}})
                _pass = True
            # FileField
            elif isinstance(field, forms.FileField):
                if initial is not None:
                    el.update({'value': None})
                el.update({
                    'view': 'uploader',
                    'autosend': False,
                    'multiple': False,
                    'directory': False,
                    'width': 100,
                    'label': _('Upload file')
                })
                if isinstance(field.widget, forms.widgets.FileInput):
                    directory = field.widget.attrs.get('directory', False)
                    multiple = field.widget.attrs.get('multiple', False),

                    el.update({
                        'multiple': field.widget.attrs.get('multiple', False),
                        'directory': field.widget.attrs.get('directory', False)
                    })

                    if directory:
                        el.update({
                            'label': _('Upload folder')
                        })

                    if multiple and not directory:
                        el.update({
                            'label': _('Upload files')
                        })

                _download = {}
                if initial:
                    _download = {
                        'name_label': name,
                        'id_label': name,
                        'borderless': True,
                        'view': "tootipButton",
                        "type": "iconButton",
                        "css": "webix_primary",
                        "icon": "fas fa-download",
                        "width": 35,
                        "on": {
                            "onItemClick": "(function (id, e) {{ window.open('{url}','_blank') }})".format(
                                url=initial.url if not isinstance(initial, six.string_types) else str(initial)
                            )
                        }
                    }
                elements.update({
                    '{}_block'.format(self[name].html_name): {
                        'cols': [
                            {
                                'name_label': name,
                                'id_label': name,
                                'borderless': True,
                                'template': label,
                                'height': 30,
                                'css': {'background-color': 'transparent !important'}
                            },
                            _download,
                            el
                        ]}})
                _pass = True
            # FilePathFied
            elif isinstance(field, forms.FilePathField):
                # el.update({'view':''})
                if initial is not None:
                    el.update({'value': initial})
            # ModelMultipleChoiceField
            elif type(field) == forms.models.ModelMultipleChoiceField:
                if settings.WEBIX_LICENSE != 'PRO':
                    raise ImproperlyConfigured(
                        _("MultipleChoiceField is available only with PRO webix license"))
                if initial is not None:
                    el.update({
                        'value': ','.join([str(i.pk) if isinstance(i, models.Model) else str(i) for i in initial])
                    })
                count = field.queryset.count()

                if count > self.min_count_suggest and \
                    name not in self.autocomplete_fields and \
                    name not in self.autocomplete_fields_exclude:
                    self.autocomplete_fields.append(name)

                # autocomplete url
                if name in self.autocomplete_fields and \
                    name not in self.autocomplete_fields_urls:
                    self.autocomplete_fields_urls.update({
                        name: self._get_url_suggest(
                            field.queryset.model._meta.app_label,
                            field.queryset.model._meta.model_name,
                            field.to_field_name
                        )
                    })

                # autocomplete field
                if name in self.autocomplete_fields:
                    el.update({
                        "view": "multicombo",
                        'selectAll': True,
                        'placeholder': _('Click to select'),
                        'options': {
                            'dynamic': True,
                            'body': {
                                'data': [],
                                'dataFeed': self.autocomplete_fields_urls[name]
                            }
                        },
                    })
                    _vals = []
                    if 'value' in el:
                        _vals = el['value'].split(",")
                    for _val in [i for i in _vals if i != ''.strip()]:
                        if field.to_field_name:
                            record = field.queryset.get(**{field.to_field_name: _val})
                            el['options']['body']['data'].append({
                                'id': '{}'.format(getattr(record, field.to_field_name)),
                                'value': '{}'.format(record)
                            })
                        else:
                            record = field.queryset.get(pk=_val)
                            el['options']['body']['data'].append({
                                'id': '{}'.format(record.pk),
                                'value': '{}'.format(record)
                            })

                # regular field without autocomplete
                else:
                    el.update({
                        "view": "multicombo",
                        'selectAll': True,
                        'placeholder': _('Click to select'),
                        'options': {
                            'dynamic': True,
                            'body': {
                                'data': self._add_null_choice([{
                                    'id': '{}'.format(i.pk),
                                    'value': '{}'.format(i)
                                } for i in field.queryset])
                            }
                        },
                    })
                    # Default if is required and there are only one option
                    if field.required and initial is None and len(field.queryset) == 1:
                        el.update({'value': '{}'.format(field.queryset.first().pk)})
            # ModelChoiceField
            elif isinstance(field, forms.models.ModelChoiceField):
                if initial is not None:
                    el.update({'value': initial.pk if isinstance(initial, models.Model) else str(initial)})
                count = field.queryset.count()

                if count > self.min_count_suggest and \
                    name not in self.autocomplete_fields and \
                    name not in self.autocomplete_fields_exclude:
                    self.autocomplete_fields.append(name)

                # autocomplete url
                if name in self.autocomplete_fields and \
                    name not in self.autocomplete_fields_urls:
                    self.autocomplete_fields_urls.update({
                        name: self._get_url_suggest(
                            field.queryset.model._meta.app_label,
                            field.queryset.model._meta.model_name,
                            field.to_field_name
                        )
                    })

                if name in self.autocomplete_fields:
                    el.update({
                        'view': 'combo',
                        'placeholder': _('Click to select'),
                        'suggest': {
                            'view': "suggest",
                            'keyPressTimeout': 400,
                            'body': {
                                'data': [],
                                'dataFeed': self.autocomplete_fields_urls[name]
                            },
                        }
                    })
                    if 'value' in el and el['value'] != '':  # and int(el['value']) > 0:
                        if field.to_field_name:
                            record = field.queryset.get(**{field.to_field_name: el['value']})
                            el['suggest']['body']['data'] = [{
                                'id': '{}'.format(getattr(record, field.to_field_name)),
                                'value': '{}'.format(record)
                            }]
                        else:
                            record = field.queryset.get(pk=el['value'])
                            el['suggest']['body']['data'] = [{
                                'id': '{}'.format(record.pk),
                                'value': '{}'.format(record)
                            }]

                # regular field without autocomplete
                elif count <= 6:
                    choices = self._add_null_choice([{
                        'id': '{}'.format(i.pk),
                        'value': '{}'.format(i)
                    } for i in field.queryset])
                    if not field.required:
                        choices.insert(0, {'id': "", 'value': "------", '$empty': True})
                    el.update({
                        'options': choices,
                        'view': 'richselect',
                        'selectAll': True,
                        'placeholder': _('Click to select')
                    })
                    # Default if is required and there are only one option
                    if field.required and initial is None and len(field.queryset) == 1:
                        el.update({'value': '{}'.format(field.queryset.first().pk)})
                else:
                    choices = self._add_null_choice([{
                        'id': '{}'.format(i.pk),
                        'value': '{}'.format(i)
                    } for i in field.queryset])
                    if not field.required:
                        choices.insert(0, {'id': "", 'value': "------", '$empty': True})
                    el.update({
                        'view': 'combo',
                        'placeholder': _('Click to select'),
                        'options': choices
                    })
            # TypedChoiceField ChoiceField
            elif isinstance(field, forms.TypedChoiceField) or isinstance(field, forms.ChoiceField):
                choices = self._add_null_choice([{
                    'id': '{}'.format(key),
                    'value': '{}'.format(value)
                } for key, value in field._choices])
                if not field.required:
                    choices.insert(0, {'id': "", 'value': "------", '$empty': True})
                el.update({
                    'selectAll': True,
                    'view': 'richselect',
                    'placeholder': _('Click to select'),
                    'options': choices
                })
                if initial is not None:
                    el.update({'value': initial})
                # Default if is required and there are only one option
                if field.required and initial is None and len(field.choices) == 1:
                    el.update({'value': '{}'.format(field.choices[0][0])})

            # JSONField (postgresql)
            elif connection.vendor == 'postgresql' and isinstance(field, JSONField):
                if isinstance(field.widget, forms.widgets.Textarea):
                    el.update({
                        'view': 'textarea'
                    })
                if initial is not None:
                    el.update({'value': dumps(initial)})

            # SimpleArrayField (postgresql)
            elif connection.vendor == 'postgresql' and \
                SimpleArrayField is not None and \
                isinstance(field, SimpleArrayField):
                if hasattr(field, 'base_field') and isinstance(field.base_field, forms.fields.DateField):
                    el.update({
                        'view': 'datepicker',
                        'multiselect': 'touch',
                        'format': "%d/%m/%Y",
                        'stringResult': True,
                    })
                    if initial is not None:
                        if isinstance(initial, list):
                            el.update({
                                'value': ', '.join([i.strftime('%Y-%m-%d') for i in initial]),
                                'orig_value': ', '.join([i.strftime('%Y-%m-%d') for i in initial]),
                            })
                elif initial is not None:
                    el.update({'value': initial})

            # CharField
            elif isinstance(field, forms.CharField):
                if isinstance(field.widget, forms.widgets.Textarea):
                    el.update({
                        'view': 'textarea'
                    })
                if initial is not None:
                    el.update({'value': initial})

            # GeoFields
            elif GeometryField is not None and isinstance(field, GeometryField):
                el.update({
                    'view': 'textarea'
                })
                if initial is not None:
                    if issubclass(GEOSGeometry, django.contrib.gis.geos.GEOSGeometry) and \
                        isinstance(initial, GEOSGeometry):
                        el.update({'value': initial.ewkt})
                    elif type(initial) == str:
                        el.update({'value': str(initial)})
                    else:
                        raise Exception(_('Initial value {} for geo field {} is not supported').format(initial, field))

            # InlineForeignKey
            elif isinstance(field, forms.models.InlineForeignKeyField):
                pass

            else:
                raise Exception(_('Type of field {} not supported').format(field))

            # widget RadioSelect
            if isinstance(field.widget, forms.RadioSelect):
                _choices = field.choices if hasattr(field, 'choices') else field.widget.choices
                el.update({
                    "view": "radio",
                    "options": [{
                        "id": key,
                        "value": value
                    } for key, value in _choices]
                })
                if initial is not None:
                    el.update({'value': initial})
                # Default if is required and there are only one option
                if field.required and initial is None and len(_choices) == 1:
                    el.update({'value': '{}'.format(_choices[0][0])})

            # widget Hidden Fields
            if type(field.widget) == forms.widgets.HiddenInput:
                el.update({'hidden': True})

            # Delete checkbox hidden
            if self.add_prefix(name).endswith('-DELETE'):
                el.update({'hidden': True})
                elements.update({
                    "{}-icon".format(self.add_prefix(name)): {
                        'view': "button",
                        'type': "icon",
                        'icon': "far fa-trash-alt",
                        'align': "center",
                        'width': 28,
                        'hidden': not (self.has_delete_permission is not None and self.has_delete_permission == True),
                        'id': "{}-icon".format(self.add_prefix(name))
                    }
                })

            if not _pass:
                elements.update({self.add_prefix(name): el})

        return elements

    @property
    def get_tabular_fields_header(self):
        """ Returns the header for the tabular inlines as webix js format """

        fields_header = OrderedDict()
        # for name, f in self.fields.items():
        for field_name, f in self.get_elements.items():

            _field_header = None

            if 'hidden' not in f or 'hidden' in f and not f['hidden']:

                _field_header = {}

                if field_name.endswith('-DELETE'):
                    _field_header = {'header': '', 'width': 28}
                else:
                    if 'width' in f:
                        _field_header.update({'width': f['width']})
                    else:
                        _field_header.update({'fillspace': True})

                if 'header' in f:
                    if not isinstance(f['header'], list):
                        f['header'] = [f['header']]
                    _field_header.update({'header': f['header']})
                elif 'label' in f:
                    _field_header.update({'header': force_text(f['label']).capitalize()})

                if _field_header is not None:
                    fields_header.update({field_name: _field_header})

        return fields_header

    @property
    def get_tabular_header(self):
        """ Returns the header for the tabular inlines as webix js format """
        return dumps({
            'id': self.add_prefix('datatable'),
            'view': "datatable",
            'scroll': 'false',
            'autoheight': 'false',
            'data': [],
            'columns': list(self.get_tabular_fields_header.values()),
        }, cls=DjangoJSONEncoder)

    def as_webix(self):
        """ Webix js format form structure """

        return dumps(list(self.get_fieldsets()), cls=DjangoJSONEncoder)[1:-1]

    def get_fieldsets(self, **kwargs):
        """ Returns a dict with all the fields """

        if 'fs' in kwargs:
            fs = kwargs['fs']
        else:
            fs = self.get_elements
        if self.style == 'tabular':
            for field in fs:
                fs[field]['label'] = ''
                fs[field]['labelWidth'] = 0
        return fs.values()


####################### BaseWebixForm AND BaseWebixModelForm #######################

class BaseWebixForm(forms.BaseForm, BaseWebixMixin):

    def __init__(self, data=None, files=None, auto_id='id_%s', prefix=None,
                 initial=None, error_class=ErrorList, label_suffix=None,
                 empty_permitted=False, field_order=None, use_required_attribute=None,
                 renderer=None, request=None, inline_id=None,
                 has_add_permission=None, has_change_permission=None, has_delete_permission=None):
        # Set request
        self.request = request
        self.has_add_permission = has_add_permission
        self.has_change_permission = has_change_permission
        self.has_delete_permission = has_delete_permission

        # Set inline id
        self.inline_id = inline_id

        # Set form prefix
        if prefix is not None:
            self.prefix = prefix

        # Update field names with prefix
        data = self.get_fields_data_with_prefix(data)

        super(BaseWebixForm, self).__init__(data, files, auto_id, prefix, initial, error_class, label_suffix,
                                            empty_permitted, field_order, use_required_attribute, renderer)

        # Set localization fields
        if settings.USE_L10N is True:
            for field in self.fields:
                self.fields[field].localize = True
                self.fields[field].widget.is_localized = True

        # TODO: check if it works with all field types on create and update action
        self.set_readonly_fields()

    def clean(self):
        cleaned_data = super(BaseWebixForm, self).clean()
        cleaned_data = self.webix_extra_clean(cleaned_data)
        return cleaned_data


class BaseWebixModelForm(forms.BaseModelForm, BaseWebixMixin):

    def __init__(self, data=None, files=None, auto_id='id_%s', prefix=None,
                 initial=None, error_class=ErrorList, label_suffix=None,
                 empty_permitted=False, instance=None, use_required_attribute=None,
                 renderer=None, request=None, inline_id=None,
                 has_add_permission=None, has_change_permission=None, has_delete_permission=None):
        # Set request
        self.request = request
        self.has_add_permission = has_add_permission
        self.has_change_permission = has_change_permission
        self.has_delete_permission = has_delete_permission

        # Set inline id
        self.inline_id = inline_id

        # Set form prefix
        if prefix is not None:
            self.prefix = prefix

        # Update field names with prefix
        data = self.get_fields_data_with_prefix(data)

        if django.VERSION > (2, 0):
            super(BaseWebixModelForm, self).__init__(data, files, auto_id, prefix, initial, error_class, label_suffix,
                                                     empty_permitted, instance, use_required_attribute, renderer)
        else:
            super(BaseWebixModelForm, self).__init__(data, files, auto_id, prefix, initial, error_class, label_suffix,
                                                     empty_permitted, instance, use_required_attribute)

        # Set localization fields
        if settings.USE_L10N is True:
            for field in self.fields:
                self.fields[field].localize = True
                self.fields[field].widget.is_localized = True

        # TODO: check if it works with all field types on create and update action
        self.set_readonly_fields()

    def clean(self):
        cleaned_data = super(BaseWebixModelForm, self).clean()
        cleaned_data = self.webix_extra_clean(cleaned_data)
        return cleaned_data

    def get_name(self):
        return capfirst(self.Meta.model._meta.verbose_name)


####################### WebixForm AND WebixModelForm #######################


class WebixForm(six.with_metaclass(DeclarativeFieldsMetaclass, BaseWebixForm)):
    pass


class WebixModelForm(six.with_metaclass(ModelFormMetaclass, BaseWebixModelForm)):
    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db.models import Q


def from_dict_to_qset(data):
    if 'qsets' in data and 'operator' in data:
        operator = data.get('operator')
        negate = data.get('negate', False)
        if len(data.get('qsets')) == 0:  # BYPASS
            qset = Q()
        for j, data_qset in enumerate(data.get('qsets')):
            if 'operator' in data_qset:
                qset_to_applicate = from_dict_to_qset(data_qset)
            else:
                qset_to_applicate = Q(**{data_qset.get('path'): data_qset.get('val')})
            if j == 0:
                qset = qset_to_applicate
            else:
                if operator == 'AND':
                    qset = qset & qset_to_applicate
                elif operator == 'OR':
                    qset = qset | qset_to_applicate
        if negate == True:
            qset = ~Q(qset)
    elif 'path' in data and 'val' in data:
        qset = Q(**{data.get('path'): data.get('val')})
    else:
        qset = Q()
    return qset



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.conf.urls import url

from django_webix.views.views_autocomplete import AutocompleteWebixLookup

urlpatterns = [
    url(r'^lookup/autocomplete/$', AutocompleteWebixLookup.as_view(), name="webix_autocomplete_lookup"),
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.template.exceptions import TemplateDoesNotExist
from django.test import TestCase
from django.urls import reverse, NoReverseMatch

from tests.app_name.models import MyModel, InlineModel, InlineStackedModel, UrlsModel


def merge_two_dicts(x, *y):
    z = x.copy()  # start with x's keys and values
    for i in y:
        z.update(i)  # modifies z with y's keys and values & returns None
    return z


class TestFakeLogin(TestCase):
    def test_login(self):
        response = self.client.get('/mylogin')
        self.assertEqual(response.status_code, 200)


class TestUrlsCalls(TestCase):
    def setUp(self):
        self.object = UrlsModel.objects.create(field='Test')

    def test_call_view_create_successurl_post(self):
        response = self.client.post('/urlsmodel/create/successurl', {'field': "Test"})
        self.assertEqual(response.status_code, 302)

    def test_call_view_create_urlcreate_post(self):
        with self.assertRaises(NoReverseMatch):
            self.client.post('/urlsmodel/create/urlcreate', {'field': "Test"})

    def test_call_view_create_urllist_post(self):
        with self.assertRaises(NoReverseMatch):
            self.client.post('/urlsmodel/create/urllist', {'field': "Test"})

    def test_call_view_create_nourl_post(self):
        with self.assertRaises(ImproperlyConfigured):
            self.client.post('/urlsmodel/create/nourl', {'field': "Test"})

    def test_call_view_update_successurl_post(self):
        response = self.client.post('/urlsmodel/update/%s/successurl' % self.object.pk, {'field': "Test"})
        self.assertEqual(response.status_code, 302)

    def test_call_view_update_urlcreate_post(self):
        with self.assertRaises(NoReverseMatch):
            self.client.post('/urlsmodel/update/%s/urlcreate' % self.object.pk, {'field': "Test"})

    def test_call_view_update_nourl_post(self):
        with self.assertRaises(ImproperlyConfigured):
            self.client.post('/urlsmodel/update/%s/nourl' % self.object.pk, {'field': "Test"})

    def test_call_view_delete_successurl_post(self):
        obj = UrlsModel.objects.create(field='Test')
        response = self.client.post('/urlsmodel/delete/%s/successurl' % obj.pk)
        self.assertEqual(response.status_code, 302)

    def test_call_view_delete_urldelete_post(self):
        obj = UrlsModel.objects.create(field='Test')
        with self.assertRaises(NoReverseMatch):
            self.client.post('/urlsmodel/delete/%s/urllist' % obj.pk)

    def test_call_view_delete_nourl_post(self):
        obj = UrlsModel.objects.create(field='Test')
        with self.assertRaises(ImproperlyConfigured):
            self.client.post('/urlsmodel/delete/%s/nourl' % obj.pk)


class TestCalls(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(username='testuser', password='12345')

        self.object_dict = {
            'field': 'Test',
            'email': 'test@test.com',
            'floatfield': 2.6,
            'decimalfield': 8.1,
            'integerfield': 1
        }

        self.object1 = MyModel.objects.create(**self.object_dict)
        self.object2 = MyModel.objects.create(**self.object_dict)
        self.inlineobject1 = InlineModel.objects.create(
            inline_field='Test',
            my_model=self.object1
        )
        self.inlineobject2 = InlineModel.objects.create(
            inline_field='Test',
            my_model=self.object2
        )

        if not InlineModel.objects.filter(pk=1).exists():
            InlineModel.objects.create(
                id=1,
                inline_field='Test',
                my_model=self.object1
            )

        for i in range(100):
            InlineStackedModel.objects.create(
                my_model=self.object2
            )

        self.object_inlines_dict = {
            'inlinemodel_set-TOTAL_FORMS': 3,
            'inlinemodel_set-INITIAL_FORMS': 0,
            'inlinemodel_set-MIN_NUM_FORMS': 0,
            'inlinemodel_set-MAX_NUM_FORMS': 1000,
            'inlinestackedmodel_set-0-booleanfield': '2',
            'inlinestackedmodel_set-0-inlinemodels': '%s,%s' % (self.inlineobject1.pk, self.inlineobject2.pk),
            'inlinestackedmodel_set-1-booleanfield': '2',
            'inlinestackedmodel_set-1-inlinemodels': '',
            'inlinestackedmodel_set-TOTAL_FORMS': 3,
            'inlinestackedmodel_set-INITIAL_FORMS': 0,
            'inlinestackedmodel_set-MIN_NUM_FORMS': 0,
            'inlinestackedmodel_set-MAX_NUM_FORMS': 1000,
            'inlineemptymodel_set-TOTAL_FORMS': 3,
            'inlineemptymodel_set-INITIAL_FORMS': 0,
            'inlineemptymodel_set-MIN_NUM_FORMS': 0,
            'inlineemptymodel_set-MAX_NUM_FORMS': 1000
        }

    def test_call_view_list(self):
        response = self.client.get('/mymodel/list')
        self.assertEqual(response.status_code, 200)

    def test_call_view_create(self):
        response = self.client.get('/mymodel/create')
        self.assertEqual(response.status_code, 200)

    def test_call_view_create_unmerged(self):
        response = self.client.get('/mymodel/create_unmerged')
        self.assertEqual(response.status_code, 200)

    def test_call_view_create_error(self):
        with self.assertRaises(TemplateDoesNotExist):
            self.client.get('/mymodel/create_error')

    def test_call_view_update(self):
        response = self.client.get('/mymodel/update/%s' % self.object1.pk)
        self.assertEqual(response.status_code, 200)

    def test_call_view_update_error(self):
        with self.assertRaises(TemplateDoesNotExist):
            self.client.get('/mymodel/update_error/%s' % self.object1.pk)

    def test_call_view_update_inlinemodel(self):
        response = self.client.get('/inlinemodel/update/%s' % self.inlineobject1.pk)
        self.assertEqual(response.status_code, 200)

    def test_call_view_delete(self):
        response = self.client.get('/mymodel/delete/%s' % self.object2.pk)
        self.assertEqual(response.status_code, 200)

    def test_call_view_delete_stacked(self):
        response = self.client.get('/inlinestackedmodel/delete/%s' % self.object2.inlinestackedmodel_set.first().pk)
        self.assertEqual(response.status_code, 200)

    def test_call_view_create_post(self):
        for i in ['loggedout', 'loggedin']:  # Logged in and logged out creation
            if i == 'loggedin':
                self.client.login(username='testuser', password='12345')
            response = self.client.post('/mymodel/create', dict(self.object_inlines_dict, **self.object_dict))
            self.assertEqual(response.status_code, 302)
            self.client.logout()
        # Verify creation by post
        pk = MyModel.objects.order_by('-id').first().pk
        self.assertEqual(MyModel.objects.get(id=pk).field, 'Test')
        response = self.client.get('/mymodel/update/%s' % pk)
        self.assertEqual(response.status_code, 200)

    def test_call_view_update_post(self):
        for i in ['loggedout', 'loggedin']:  # Logged in and logged out creation
            if i == 'loggedin':
                self.client.login(username='testuser', password='12345')
            response = self.client.post(
                '/mymodel/update/%s' % self.object1.pk,
                merge_two_dicts(self.object_inlines_dict, self.object_dict, {'id': self.object1.pk})
            )
            self.assertEqual(response.status_code, 302)
            self.client.logout()
        # Verify creation by post
        self.assertEqual(MyModel.objects.get(id=self.object1.pk).field, 'Test')

    def test_call_view_errors_update_post(self):
        self.client.login(username='testuser', password='12345')
        with self.assertRaises(ValidationError):
            self.client.post('/mymodel/update/%s' % self.object1.pk, {
                'inlinestackedmodel_set-INITIAL_FORMS': 3,
                'inlinestackedmodel_set-MIN_NUM_FORMS': 1,
                'inlinestackedmodel_set-MAX_NUM_FORMS': 1000
            })
        # self.assertEqual(response.status_code, 200)
        self.client.logout()

    def test_call_view_delete_post(self):
        for i in ['loggedout', 'loggedin']:  # Logged in and logged out update
            if i == 'loggedin':
                self.client.login(username='testuser', password='12345')
                response = self.client.post('/mymodel/delete/%s' % self.object2.pk, {})
            else:
                response = self.client.post('/mymodel/delete/%s' % self.object1.pk, {})
            self.assertEqual(response.status_code, 302)
            self.client.logout()

    def test_autocomplete_logged_querysetring(self):
        self.client.login(username='testuser', password='12345')
        url = "%(url)s?app_label=%(app_label)s&model_name=%(model_name)s&to_field=%(to_field)s&" \
              "filter[value]=%(value)s&query_string=%(query_string)s"
        response = self.client.get(url % {
            'url': reverse('webix_autocomplete_lookup'),
            'model_name': 'mymodel',
            'app_label': 'app_name',
            'to_field': 'id',
            'value': 'T',
            'query_string': 'field=T|field=E:field=T|boolean=TRUE|boolean=FALSE|_to_field=test|field=None'
        })
        self.assertEqual(response.status_code, 200)

    def test_autocomplete_without_querysetring(self):
        for i in ['loggedout', 'loggedin']:  # Logged in and logged out update
            if i == 'loggedin':
                self.client.login(username='testuser', password='12345')
            url = "%(url)s?app_label=%(app_label)s&model_name=%(model_name)s&to_field=%(to_field)s&" \
                  "filter[value]=%(value)s&nolimit=True"
            response = self.client.get(url % {
                'url': reverse('webix_autocomplete_lookup'),
                'model_name': 'mymodel',
                'app_label': 'app_name',
                'to_field': 'id',
                'value': 'T'
            })
            if i == 'loggedout':
                self.assertEqual(response.status_code, 403)
            else:
                self.assertEqual(response.status_code, 200)
            self.client.logout()

    def test_autocomplete_model_without_autocomplete(self):
        self.client.login(username='testuser', password='12345')
        url = "%(url)s?app_label=%(app_label)s&model_name=%(model_name)s&to_field=%(to_field)s&filter[value]=%(value)s"
        response = self.client.get(url % {
            'url': reverse('webix_autocomplete_lookup'),
            'model_name': 'inlinestackedmodel',
            'app_label': 'app_name',
            'to_field': 'id',
            'value': 'T'
        })
        self.assertEqual(response.status_code, 200)

    def test_autocomplete_model_foreignkey(self):
        self.client.login(username='testuser', password='12345')
        url = "%(url)s?app_label=%(app_label)s&model_name=%(model_name)s&to_field=%(to_field)s&filter[value]=%(value)s"
        response = self.client.get(url % {
            'url': reverse('webix_autocomplete_lookup'),
            'model_name': 'inlinemodel',
            'app_label': 'app_name',
            'to_field': 'id',
            'value': 'T'
        })
        self.assertEqual(response.status_code, 200)

    def test_autocomplete_invalid(self):
        self.client.login(username='testuser', password='12345')
        url = "%(url)s?app_label=%(app_label)s&to_field=%(to_field)s&filter[value]=%(value)s"
        response = self.client.get(url % {
            'url': reverse('webix_autocomplete_lookup'),
            'app_label': 'app_name',
            'to_field': 'id',
            'value': 'T'
        })
        self.assertEqual(response.status_code, 200)

    def test_autocomplete_error(self):
        self.client.login(username='testuser', password='12345')
        url = "%(url)s?app_label=%(app_label)s&model_name=%(model_name)s&to_field=%(to_field)s&filter[value]=%(value)s"
        response = self.client.get(url % {
            'url': reverse('webix_autocomplete_lookup'),
            'model_name': 'fakemodel',
            'app_label': 'app_name',
            'to_field': 'id',
            'value': 'T'
        })
        self.assertEqual(response.status_code, 200)



from django_webix.admin_webix.decorators import register
from django_webix.admin_webix.options import ModelWebixAdmin
from django_webix.admin_webix.sites import AdminWebixSite, site
from django.utils.module_loading import autodiscover_modules

__all__ = [
    "register", "ModelWebixAdmin", "AdminWebixSite", "site", "autodiscover",
]


def autodiscover():
    autodiscover_modules('admin_webix', register_to=site)


default_app_config = 'django_webix.admin_webix.apps.AdminWebixConfig'



# -*- coding: utf-8

from __future__ import unicode_literals, absolute_import

import os

import django

DEBUG = True
USE_TZ = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "=================================================="

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': 'djangowebix',
        'USER': 'postgres',
        'HOST': 'localhost',
        'PORT': '5432',
        'TEST': {
            'NAME':  os.environ.get('DATABASE_NAME', 'djangowebix_test'),
        },
    }
}

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.sites",

    "sorl.thumbnail",

    "django_webix",
    "tests.app_name"
]

ROOT_URLCONF = 'tests.urls'

SITE_ID = 1

if django.VERSION >= (1, 10):
    MIDDLEWARE = (
        'django.middleware.security.SecurityMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    )
else:
    MIDDLEWARE_CLASSES = (
        'django.middleware.security.SecurityMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    )

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.i18n',
            ],
        },
    },
]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.core.exceptions import PermissionDenied
from django.utils.translation import ugettext as _
from django.views.generic import DetailView

from django_webix.views.generic.base import WebixBaseMixin, WebixPermissionsMixin, WebixUrlMixin


class WebixDetailView(WebixBaseMixin, WebixPermissionsMixin, WebixUrlMixin, DetailView):
    template_name = 'django_webix/generic/detail.js'

    def get_object(self, queryset=None):
        if getattr(self, 'object', None) is not None:
            return self.object
        return super(WebixDetailView, self).get_object(queryset=queryset)

    def dispatch(self, *args, **kwargs):
        self.object = self.get_object()
        if not self.has_view_permission(request=self.request, obj=self.object):
            raise PermissionDenied(_('View permission is not allowed'))
        return super(WebixDetailView, self).dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(WebixDetailView, self).get_context_data(**kwargs)
        context.update(self.get_context_data_webix_permissions(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_url(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_base(request=self.request))
        return context



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from functools import wraps
from django.conf import settings
from django.shortcuts import render

def script_login_required(view, login_url=None):
    @wraps(view)
    def wrapper(request, *args, **kwargs):
        if not request.user.is_authenticated:
            return render(request, 'django_webix/redirect_to_login.js', context={
                'login_url': login_url or settings.LOGIN_URL
            })
        return view(request, *args, **kwargs)

    return wrapper



from django.http import JsonResponse
from django.utils.translation import ugettext as _
from django_webix.views.generic.decorators import action_config


@action_config(action_key='delete',
               response_type='json',
               short_description=_('Delete'),
               allowed_permissions=['delete'])
def multiple_delete_action(self, request, qs):
    _count_delete_instances = int(qs.count())
    qs.delete()
    return JsonResponse({
        "status": True,
        "message": _('{count_delete_instances} elements have been deleted').format(
            count_delete_instances=_count_delete_instances),
        "redirect_url": self.get_url_list(),
    }, safe=False)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import types

from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.text import capfirst
from django.contrib.admin.utils import NestedObjects


def walk_items(item_list):
    item_iterator = iter(item_list)
    try:
        item = next(item_iterator)
        while True:
            try:
                next_item = next(item_iterator)
            except StopIteration:
                yield item, None
                break
            if isinstance(next_item, (list, tuple, types.GeneratorType)):
                try:
                    iter(next_item)
                except TypeError:  # pragma: no cover
                    pass
                else:
                    yield item, next_item
                    item = next(item_iterator)
                    continue  # pragma: no cover
            yield item, None
            item = next_item
    except StopIteration:
        pass


def tree_formatter(item_list):
    output = []
    for item, children in walk_items(item_list):
        _data = {"value": '{}: {}'.format(capfirst(item._meta.verbose_name), force_text(item))}
        _sub_data = None
        if children:
            _data.update({"data": tree_formatter(children), "open": True})
        output.append(_data)
    return output


class NestedObjectsWithLimit(NestedObjects):
    exclude_models = None
    only_models = None

    def __init__(self, *args, **kwargs):
        self.exclude_models = kwargs.pop('exclude_models', None)
        self.only_models = kwargs.pop('only_models', None)
        if self.exclude_models is not None and self.only_models is not None:
            raise Exception('Set only one of exclude_models or only_models parameter')
        super().__init__(*args, **kwargs)


    def add_edge(self, source, target):
        if self.only_models is None or (source is None or source._meta.model in self.only_models):
            if self.only_models is None or (target is not None and target._meta.model in self.only_models):
                super(NestedObjectsWithLimit, self).add_edge(source, target)
#            else:
#                super(NestedObjectsWithLimit, self).add_edge(source, None)
        elif self.exclude_models is None or (source is None or source._meta.model not in self.exclude_models):
            if self.exclude_models is None or (target is not None and target._meta.model not in self.only_models):
                super(NestedObjectsWithLimit, self).add_edge(source, target)
#            else:
#                super(NestedObjectsWithLimit, self).add_edge(source, None)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import copy
import json

from django.apps import apps
from django.conf import settings

from django.core.exceptions import ImproperlyConfigured
from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _
from django.views.generic import DeleteView

from django_webix.views.generic.base import WebixBaseMixin, WebixPermissionsMixin, WebixUrlMixin
from django_webix.views.generic.signals import django_webix_view_pre_delete, django_webix_view_post_delete
from django_webix.views.generic.utils import tree_formatter, NestedObjectsWithLimit


class WebixDeleteView(WebixBaseMixin, WebixPermissionsMixin, WebixUrlMixin, DeleteView):
    logs_enable = True
    exclude_models = None
    only_models = None

    template_name = 'django_webix/generic/delete.js'

    def get_exclude_models(self):
        return self.exclude_models

    def get_only_models(self):
        return self.only_models

    def get_object(self, queryset=None):
        if getattr(self, 'object', None) is not None:
            return self.object
        return super(WebixDeleteView, self).get_object(queryset=queryset)

    def dispatch(self, *args, **kwargs):
        self.object = self.get_object()
        if self.request.method == 'GET':
            if not self.has_view_permission(request=self.request, obj=self.object):
                raise PermissionDenied(_('View permission is not allowed'))
        else:
            if not self.has_delete_permission(request=self.request, obj=self.object):
                raise PermissionDenied(_('Delete permission is not allowed'))
        return super(WebixDeleteView, self).dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(WebixDeleteView, self).get_context_data(**kwargs)
        context.update(self.get_context_data_webix_permissions(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_url(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_base(request=self.request))
        context.update({'multiple_delete_confirmation': getattr(settings, 'WEBIX_MULTIPLE_DELETE_CONFIRMATION', True)})
        # Nested objects
        collector = NestedObjectsWithLimit(using='default',
                                           exclude_models=self.get_exclude_models(),
                                           only_models=self.get_only_models())
        collector.collect([self.object])
        context.update({
            'related_objects': json.dumps(tree_formatter(collector.nested())),
        })

        return context

    def get_success_url(self):
        if self.success_url is not None:
            url = self.success_url
        elif self.get_url_list() is not None:
            url = self.get_url_list()
        else:
            raise ImproperlyConfigured(_(
                "No URL to redirect to.  Either provide a url or define"
                " a get_absolute_url method on the Model."))
        return url

    def get_failure_url(self):
        if self.failure_url:
            url = self.failure_url
        elif self.get_delete(obj=self.object) is not None:
            url = self.get_delete(obj=self.object)
        else:
            raise ImproperlyConfigured(_(
                "No URL to failure redirect to.  Either provide a url or define"
                " a get_absolute_url method on the Model."))
        return url

    def pre_delete_valid(self, **kwargs):
        django_webix_view_pre_delete.send(sender=self, instance=self.object)

    def post_delete_valid(self, **kwargs):
        django_webix_view_post_delete.send(sender=self, instance=self.copied_object)

    def response_valid(self, success_url=None, **kwargs):
        return HttpResponseRedirect(success_url)

    def response_invalid(self, failure_url=None, **kwargs):
        return HttpResponseRedirect(failure_url)

    def delete(self, request, *args, **kwargs):

        self.object = self.get_object()
        self.copied_object = copy.deepcopy(self.object)

        anonymous = request.user.is_anonymous() if callable(request.user.is_anonymous) else request.user.is_anonymous
        if len(self.get_failure_delete_related_objects(request=request, obj=self.object)) > 0:
            return self.response_invalid(failure_url=self.get_failure_url())

        if self.logs_enable is True and not anonymous and apps.is_installed('django.contrib.admin'):
            from django.contrib.admin.models import LogEntry, DELETION
            from django.contrib.contenttypes.models import ContentType
            LogEntry.objects.log_action(
                user_id=request.user.id,
                content_type_id=ContentType.objects.get_for_model(self.object).pk,
                object_id=self.object.pk,
                object_repr=force_text(self.object),
                action_flag=DELETION
            )

        self.pre_delete_valid()

        success_url = self.get_success_url()
        self.object.delete()

        self.post_delete_valid()

        return self.response_valid(success_url=success_url)



from django.conf.urls import url, include

from tests.app_name.views import MyLoginView
from tests.app_name.views import MyModelListView, MyModelCreateView, MyModelCreateUnmergedView, \
    MyModelCreateErrorView, MyModelUpdateView, MyModelUpdateErrorView, MyModelDeleteView
from tests.app_name.views import CreateSuccessUrlView, CreateUrlUpdateView, CreateUrlListView, CreateNoUrlView
from tests.app_name.views import UpdateSuccessUrlView, UpdateUrlUpdateView, UpdateNoUrlView
from tests.app_name.views import DeleteSuccessUrlView, DeleteUrlListView, DeleteNoUrlView

from tests.app_name.views import InlineModelUpdateView
from tests.app_name.views import InlineStackedModelDelete

urlpatterns = [
    url(r'^django-webix/', include('django_webix.urls')),

    url(r'^mylogin$', MyLoginView.as_view(), name='mylogin'),

    url(r'^mymodel/list$', MyModelListView.as_view(), name='app_name.mymodel.list'),
    url(r'^mymodel/create$', MyModelCreateView.as_view(), name='app_name.mymodel.create'),
    url(r'^mymodel/create_unmerged$', MyModelCreateUnmergedView.as_view(), name='app_name.mymodel.create_unmerged'),
    url(r'^mymodel/create_error$', MyModelCreateErrorView.as_view(), name='app_name.mymodel.create_error'),
    url(r'^mymodel/update/(?P<pk>\d+)$', MyModelUpdateView.as_view(), name='app_name.mymodel.update'),
    url(r'^mymodel/update_error/(?P<pk>\d+)$', MyModelUpdateErrorView.as_view(), name='app_name.mymodel.update_error'),
    url(r'^mymodel/delete/(?P<pk>\d+)$', MyModelDeleteView.as_view(), name='app_name.mymodel.delete'),

    url(r'^inlinemodel/update/(?P<pk>\d+)$', InlineModelUpdateView.as_view(), name='app_name.inlinemodel.update'),

    url(r'^inlinestackedmodel/delete/(?P<pk>\d+)$', InlineStackedModelDelete.as_view(),
        name='app_name.inlinestackedmodel.delete'),

    url(r'^urlsmodel/create/successurl$', CreateSuccessUrlView.as_view()),
    url(r'^urlsmodel/create/urlcreate$', CreateUrlUpdateView.as_view()),
    url(r'^urlsmodel/create/urllist$', CreateUrlListView.as_view()),
    url(r'^urlsmodel/create/nourl$', CreateNoUrlView.as_view()),

    url(r'^urlsmodel/update/(?P<pk>\d+)/successurl$', UpdateSuccessUrlView.as_view()),
    url(r'^urlsmodel/update/(?P<pk>\d+)/urlcreate$', UpdateUrlUpdateView.as_view()),
    url(r'^urlsmodel/update/(?P<pk>\d+)/nourl$', UpdateNoUrlView.as_view()),

    url(r'^urlsmodel/delete/(?P<pk>\d+)/successurl$', DeleteSuccessUrlView.as_view()),
    url(r'^urlsmodel/delete/(?P<pk>\d+)/urllist$', DeleteUrlListView.as_view()),
    url(r'^urlsmodel/delete/(?P<pk>\d+)/nourl$', DeleteNoUrlView.as_view()),
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import models
from django.utils import timezone



class MyModel(models.Model):
    field = models.CharField(max_length=255)
    email = models.EmailField(max_length=255)
    floatfield = models.FloatField()
    decimalfield = models.DecimalField(decimal_places=2, max_digits=5)
    integerfield = models.IntegerField()
    integerfield_null = models.IntegerField(blank=True, null=True)
    integerfield_default = models.IntegerField(blank=True, null=True, default=43)
    readonly = models.CharField(max_length=255, blank=True, null=True)
    boolean = models.NullBooleanField(blank=True, null=True, default=True)
    datefield = models.DateField(blank=True, null=True, default=timezone.datetime.today)
    datefield_empty = models.DateField(blank=True, null=True)
    datefield_2 = models.DateField(blank=True, null=True, default="2018-01-01")
    datetimefield = models.DateTimeField(blank=True, null=True, default=timezone.now)
    datetimefield_empty = models.DateTimeField(blank=True, null=True)
    datetimefield_2 = models.DateTimeField(blank=True, null=True, default="2018-01-01T13:00:00+00:00")
    booleanfield = models.BooleanField(default=False)
    datetimefield_null = models.DateTimeField(blank=True, null=True)
    filepathfield_default = models.FilePathField(blank=True, null=True, path='/tmp', default='')

    class Meta:
        ordering = ['field']

    @staticmethod
    def autocomplete_search_fields():
        return "field__icontains", "email__icontains",


class InlineModel(models.Model):
    inline_field = models.CharField(max_length=255)
    timefield = models.TimeField(blank=True, null=True, default=timezone.now)
    datefield = models.DateField(blank=True, null=True, default=timezone.datetime.today)
    datetimefield = models.DateTimeField(blank=True, null=True, default=timezone.now)
    filepathfield = models.FilePathField(blank=True, null=True, path='/tmp')
    filefield = models.FileField(blank=True, null=True)
    filefield_default = models.FileField(blank=True, null=True, default='test.txt')
    imagefield = models.ImageField(blank=True, null=True)
    imagefield_default = models.ImageField(default='image.jpg')
    booleanfield = models.BooleanField(default=False)

    my_model = models.ForeignKey(MyModel, on_delete=models.CASCADE)

    class Meta:
        ordering = ['my_model__field', 'my_model']


class InlineStackedModel(models.Model):
    textfield = models.TextField(blank=True, null=True)
    timefield = models.TimeField(blank=True, null=True)
    timefield_string = models.TimeField(blank=True, null=True, default="10:00")  # TODO: SISTEMARE
    datefield = models.DateField(blank=True, null=True)
    nullbooleanfield = models.NullBooleanField(blank=True, null=True)
    booleanfield = models.BooleanField(default=False)
    urlfield = models.URLField(blank=True, null=True, default='https://www.example.com')
    urlfield_2 = models.URLField(blank=True, null=True)
    slugfield = models.SlugField(blank=True, null=True, default='slug-field')
    slugfield_2 = models.SlugField(blank=True, null=True)
    choicefield = models.CharField(max_length=255, blank=True, null=True, choices=(
        ('test', 'Test'),
        ('sample', 'Sample')
    ), default='test')
    inlinemodel = models.ForeignKey('InlineModel', blank=True, null=True, related_name='inlinemodel_fk',
                                    on_delete=models.CASCADE)
    inlinemodels = models.ManyToManyField('InlineModel', blank=True, related_name='inlinemodel_m2m')
    inlinemodels_default = models.ManyToManyField('InlineModel', blank=True, related_name='inlinemodel_m2m_default',
                                                  default=[1])
    foreign_default = models.ForeignKey('InlineModel', blank=True, null=True, default=1, on_delete=models.CASCADE)

    my_model = models.ForeignKey(MyModel, on_delete=models.CASCADE)


class InlineEmptyModel(models.Model):
    textfield = models.TextField(blank=True, null=True)
    boolean = models.NullBooleanField(blank=True, null=True, default=2)
    choicefield = models.CharField(max_length=255, choices=(
        ('test', 'Test'),
        ('sample', 'Sample')
    ))
    foreign_default = models.ForeignKey('InlineStackedModel', on_delete=models.CASCADE)
    foreign_default_2 = models.ForeignKey('InlineStackedModel', blank=True, null=True,
                                          related_name="inlineeptymodel_fk2", on_delete=models.CASCADE)
    foreign_self = models.ForeignKey('self', on_delete=models.CASCADE)

    my_model = models.ForeignKey(MyModel, on_delete=models.CASCADE)


class SubInlineStackedModel(models.Model):
    my_model = models.ForeignKey(InlineStackedModel, on_delete=models.CASCADE)


class UrlsModel(models.Model):
    field = models.CharField(max_length=255)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import six
from django.conf import settings
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView
from django.urls import reverse

from django.urls.exceptions import NoReverseMatch

class WebixPermissionsMixin:
    model = None
    request = None
    check_permissions = True

    add_permission = None
    change_permission = None
    delete_permission = None
    view_permission = None
    view_or_change_permission = None
    module_permission = None

    _failure_add_related_objects = None
    _failure_change_related_objects = None
    _failure_delete_related_objects = None
    _failure_view_related_objects = None

    remove_disabled_buttons = False

    def get_failure_add_missing_objects(self, request):
        return [
#            {
#                'text': '',
#                'url': '',
#            },
        ]

    def get_failure_add_related_objects(self, request):
        return []

    def get_failure_change_related_objects(self, request, obj=None):
        return []

    def get_failure_delete_related_objects(self, request, obj=None):
        return []

    def get_failure_view_related_objects(self, request, obj=None):
        return []

    def has_add_django_user_permission(self, user):
        if self.model is not None:
            return user.has_perm('{}.add_{}'.format(self.model._meta.app_label, self.model._meta.model_name))
        return False

    def has_change_django_user_permission(self, user):
        if self.model is not None:
            return user.has_perm('{}.change_{}'.format(self.model._meta.app_label, self.model._meta.model_name))
        return False

    def has_delete_django_user_permission(self, user):
        if self.model is not None:
            return user.has_perm('{}.delete_{}'.format(self.model._meta.app_label, self.model._meta.model_name))
        return False

    def has_view_django_user_permission(self, user):
        if self.model is not None:
            return user.has_perm('{}.view_{}'.format(self.model._meta.app_label, self.model._meta.model_name))
        return False

    def has_add_permission(self, request):
        if not self.check_permissions:
            return True
        if self.add_permission is not None:
            return self.add_permission
        if len(self.get_failure_add_related_objects(request)) > 0:
            return False
        return self.has_add_django_user_permission(user=request.user)

    def has_change_permission(self, request, obj=None):
        if not self.check_permissions:
            return True
        if self.change_permission is not None:
            return self.change_permission
        if len(self.get_failure_change_related_objects(request, obj=obj)) > 0:
            return False
        return self.has_change_django_user_permission(user=request.user)

    def has_delete_permission(self, request, obj=None):
        if not self.check_permissions:
            return True
        if self.delete_permission is not None:
            return self.delete_permission
        if len(self.get_failure_delete_related_objects(request, obj=obj)) > 0:
            return False
        return self.has_change_django_user_permission(user=request.user)

    def has_view_permission(self, request, obj=None):
        if not self.check_permissions:
            return True
        if self.view_permission is not None:
            return self.view_permission
        if len(self.get_failure_view_related_objects(request, obj=obj)) > 0:
            return False
        return self.has_view_django_user_permission(user=request.user)

    def get_info_no_add_permission(self, has_permission, request):
        if not has_permission:
            return [_("You haven't add permission")]
        return []

    def get_info_no_change_permission(self, has_permission, request, obj=None):
        if not has_permission:
            return [_("You haven't change permission")]
        return []

    def get_info_no_delete_permission(self, has_permission, request, obj=None):
        if not has_permission:
            return [_("You haven't delete permission")]
        return []

    def get_info_no_view_permission(self, has_permission, request, obj=None):
        if not has_permission:
            return [_("You haven't view permission")]
        return []

    def has_view_or_change_permission(self, request, obj=None):
        if self.view_permission is not None or self.change_permission is not None:
            return self.view_permission or self.change_permission
        return self.has_view_permission(request=request, obj=obj) or \
               self.has_change_permission(request=request, obj=obj)

    def has_module_permission(self, request):
        if not self.check_permissions:
            return True
        if self.module_permission is not None:
            return self.module_permission
        if self.model is not None:
            return request.user.has_module_perms(self.model._meta.app_label)
        return False

    def get_context_data_webix_permissions(self, request, obj=None, **kwargs):
        _has_view_permission = self.has_view_permission(request=self.request, obj=obj)
        _has_add_permission = self.has_add_permission(request=self.request)
        _has_change_permission = self.has_change_permission(request=self.request, obj=obj)
        _has_delete_permission = self.has_delete_permission(request=self.request, obj=obj)
        return {
            # Buttons
            'remove_disabled_buttons': self.remove_disabled_buttons,
            # Permissions
            'has_view_permission': _has_view_permission,
            'has_add_permission': _has_add_permission,
            'has_change_permission': _has_change_permission,
            'has_delete_permission': _has_delete_permission,
            'has_view_or_change_permission': self.has_view_or_change_permission(request=self.request, obj=obj),
            'has_module_permission': self.has_module_permission(request=self.request),
            # info no permissions
            'info_no_add_permission': self.get_info_no_add_permission(has_permission=_has_add_permission,
                                                                      request=self.request),
            'info_no_change_permission': self.get_info_no_change_permission(has_permission=_has_change_permission,
                                                                            request=self.request,
                                                                            obj=obj),
            'info_no_delete_permission': self.get_info_no_delete_permission(has_permission=_has_delete_permission,
                                                                            request=self.request,
                                                                            obj=obj),
            'info_no_view_permission': self.get_info_no_view_permission(has_permission=_has_view_permission,
                                                                        request=self.request,
                                                                        obj=obj),
            # failure related objects
            'failure_view_related_objects': self.get_failure_view_related_objects(request=self.request,
                                                                                  obj=obj),
            'failure_add_related_objects': self.get_failure_add_related_objects(request=self.request),
            'failure_change_related_objects': self.get_failure_change_related_objects(request=self.request,
                                                                                      obj=obj),
            'failure_delete_related_objects': self.get_failure_delete_related_objects(request=self.request,
                                                                                      obj=obj),
            # filure add missing_objects
            'failure_add_missing_objects': self.get_failure_add_missing_objects(request=self.request),
        }


class WebixUrlMixin:
    model = None

    url_pattern_list = None
    url_pattern_create = None
    url_pattern_update = None
    url_pattern_delete = None

    def is_popup(self):
        return self.request.GET.get('_popup', self.request.POST.get('_popup', False)) != False

    def wrap_url_popup(self, url):
        if url is not None:
            if self.is_popup():
                if '?' in url:
                    return url + '&_popup'
                else:
                    return url + '?_popup'
            else:
                return url
        else:
            return None

    def get_url_pattern_list(self):
        if self.url_pattern_list is not None:
            return self.url_pattern_list
        else:
            return '{}.list'.format(self.get_model_name())

    def get_url_pattern_create(self):
        if self.url_pattern_create is not None:
            return self.url_pattern_create
        else:
            return '{}.create'.format(self.get_model_name())

    def get_url_pattern_update(self):
        if self.url_pattern_update is not None:
            return self.url_pattern_update
        else:
            return '{}.update'.format(self.get_model_name())

    def get_url_pattern_delete(self):
        if self.url_pattern_delete is not None:
            return self.url_pattern_delete
        else:
            return '{}.delete'.format(self.get_model_name())

    def _check_url(self, url_name, reverse_kwargs={}):
        """
        Check if url_name exists

        :param url_name: url name
        :return: url_name if exists, otherwhise None
        """
        try:
            url = reverse(url_name, kwargs=reverse_kwargs)
            return url_name
        except NoReverseMatch:
            return None

    def get_model_name(self):
        if self.model is not None:
            return '{}.{}'.format(self.model._meta.app_label, self.model._meta.model_name)
        return None

    def get_url_list(self):
        if self.model is not None:
            _url_pattern_name = self._check_url(self.get_url_pattern_list())
            if _url_pattern_name is not None:
                return self.wrap_url_popup( reverse(_url_pattern_name) )
        return None

    def get_url_create_kwargs(self):
        return None

    def get_url_create(self):
        if self.model is not None:
            create_kwargs = self.get_url_create_kwargs()
            if create_kwargs is not None:
                _url_pattern_name = self._check_url(self.get_url_pattern_create(),reverse_kwargs=self.get_url_create_kwargs())
            else:
                _url_pattern_name = self._check_url(self.get_url_pattern_create())
            if _url_pattern_name is not None:
                if create_kwargs is not None:
                    return self.wrap_url_popup( reverse(_url_pattern_name, kwargs=create_kwargs) )
                else:
                    return self.wrap_url_popup( reverse(_url_pattern_name) )

        return None

    def get_url_update(self, obj=None):
        if self.model is not None:
            _url_pattern_name = self._check_url(self.get_url_pattern_update(), {'pk':0})
            if _url_pattern_name is not None:
                if obj is not None:
                    _pk = obj.pk
                else:
                    _pk = 0
                return self.wrap_url_popup( reverse(_url_pattern_name, kwargs={'pk': _pk}) )
        return None

    def get_url_delete(self, obj=None):
        if self.model is not None:
            _url_pattern_name = self._check_url(self.get_url_pattern_delete(), {'pk':0})
            if _url_pattern_name is not None:
                if obj is not None:
                    _pk = obj.pk
                else:
                    _pk = 0
                return self.wrap_url_popup( reverse(_url_pattern_name, kwargs={'pk': _pk}) )
        return None

    def get_view_prefix(self):
        return '{}_{}_'.format(self.model._meta.app_label, self.model._meta.model_name)

    def get_context_data_webix_url(self, request, obj=None, **kwargs):
        return {
            # Urls
            'url_list': self.get_url_list(),
            'url_create': self.get_url_create(),
            'url_update': self.get_url_update(obj=obj),
            'url_delete': self.get_url_delete(obj=obj),
            # Model info
            'is_popup': self.is_popup(),
            'model': self.model,
            'model_name': self.get_model_name(),
            'app_label': self.model._meta.app_label if self.model else None,
            'module_name': self.model._meta.model_name if self.model else None,
            'view_prefix': self.get_view_prefix()
        }


class WebixBaseMixin:

    def get_container_id(self, request):
        return settings.WEBIX_CONTAINER_ID

    def get_overlay_container_id(self, request):
        return getattr(settings, 'WEBIX_OVERLAY_CONTAINER_ID', settings.WEBIX_CONTAINER_ID)

    def get_context_data_webix_base(self, request, **kwargs):
        return {
            'webix_container_id': self.get_container_id(request=self.request),
            'webix_overlay_container_id': self.get_overlay_container_id(request=self.request)
        }


class WebixTemplateView(WebixBaseMixin, TemplateView):

    def get_context_data(self, **kwargs):
        context = super(WebixTemplateView, self).get_context_data(**kwargs)
        context.update(self.get_context_data_webix_base(request=self.request))
        return context

    def get_view_prefix(self):
        return ''



import os
import re
import sys

from setuptools import find_packages

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    """Retrieves the version from django_dandelion/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version("django_webix", "__init__.py")

if sys.argv[-1] == 'publish':
    try:
        import wheel

        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

readme = open('README.rst').read()
changelog = open('CHANGELOG.rst').read().replace('.. :changelog:', '')

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-webix',
    version=version,
    packages=find_packages(),
    include_package_data=True,
    license='GNU GPLv3 License',
    description='',
    long_description=readme,
    url='https://github.com/MPASolutions/django-webix',
    author='MPA Solutions',
    author_email='info@mpasol.it',
    install_requires=[
        'Django>=1.11',
        'sorl-thumbnail>=12.4.1',
        'django-extra-views>=0.11.0',
        'six'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',  # example license
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.apps import AppConfig
from django.conf import settings


class DjangoWebixConfig(AppConfig):
    name = 'django_webix'

    def ready(self):
        # check webix version
        if not hasattr(settings, 'WEBIX_VERSION'):
            raise Exception('WEBIX_VERSION is not found in your settings (eg.6.4.1)')

        # check webix license (we have to disable some form controls)
        # in accord to https://docs.webix.com/desktop__controls.html
        if not hasattr(settings, 'WEBIX_LICENSE'):
            raise Exception('WEBIX_LICENSE is not found in your settings ("FREE" or "PRO")')

        # check webix container id
        if not hasattr(settings, 'WEBIX_CONTAINER_ID'):
            raise Exception('WEBIX_CONTAINER_ID is webix_id where js interfaces are loaded and '
                            'is not found in your settings (usually "content_right")')



from django_webix.views.generic.base import WebixTemplateView
from django_webix.views.generic.detail import WebixDetailView
from django_webix.views.generic.list import WebixListView, WebixTemplateListView
from django_webix.views.generic.create_update import (
    WebixCreateView, WebixUpdateView, WebixCreateWithInlinesView, WebixCreateWithInlinesUnmergedView,
    WebixUpdateWithInlinesView, WebixUpdateWithInlinesUnmergedView
)
from django_webix.views.generic.delete import WebixDeleteView

__all__ = [
    # model views
    'WebixCreateView', 'WebixUpdateView', 'WebixDetailView', 'WebixListView', 'WebixDeleteView',
    # template views
    'WebixTemplateView', 'WebixTemplateListView',
    # deprecated
    'WebixCreateWithInlinesView', 'WebixCreateWithInlinesUnmergedView', 'WebixUpdateWithInlinesView',
    'WebixUpdateWithInlinesUnmergedView'
]



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import json

import django
from django.apps import apps
from django.core.exceptions import FieldDoesNotExist
from django.core.exceptions import PermissionDenied, ImproperlyConfigured
from django.db.models import Q, F, ManyToManyField
from django.http import JsonResponse, Http404
from django.template import Template, Context
from django.template.loader import get_template
from django.utils.translation import ugettext as _
from django.views.generic import ListView

from django_webix.utils.filters import from_dict_to_qset
from django_webix.views.generic.base import WebixBaseMixin, WebixPermissionsMixin, WebixUrlMixin

try:
    from django.contrib.gis.geos import GEOSGeometry
except ImportError:
    GEOSGeometry = object

try:
    from django.contrib.gis.geos import MultiPolygon
except ImportError:
    MultiPolygon = object

class WebixListView(WebixBaseMixin,
                    WebixPermissionsMixin,
                    WebixUrlMixin,
                    ListView):
    # request vars
    http_method_names = ['get', 'post']  # enable POST for filter porpouse

    # queryset vars
    pk_field = None
    fields = None  # ex. [{'field_name':'XXX','datalist_column':'YYY',}]
    order_by = None

    actions = []  # [multiple_delete_action]

    # paging
    enable_json_loading = False
    paginate_count_default = 100
    paginate_start_default = 0
    paginate_count_key = 'count'
    paginate_start_key = 'start'

    # template vars
    template_name = 'django_webix/generic/list.js'
    title = None
    actions_style = None
    enable_column_copy = True
    enable_column_delete = True
    enable_row_click = True
    type_row_click = 'single'  # or 'double'
    enable_actions = True

    def is_installed_django_webix_filter(self):
        return apps.is_installed('django_webix_filter')

    ########### QUERYSET BUILDER ###########

    # {
    #     'operator': 'AND', # by default
    #     'qsets':[
    #         {'path': 'appezzamento__superficie__in','val': [1,2],}
    #         {'geo_field_name': '...','geo_srid': '...','polygons':'...'}
    #     ]
    # }
    def _decode_text_filters(self, request, key):
        if request.method == 'GET':
            filters_text = request.GET.get(key, None)
        elif request.method == 'POST':
            filters_text = request.POST.get(key, None)
        else:
            filters_text = None

        filters = None
        if filters_text is not None:
            if type(filters_text) == str:
                # NB: JSON syntax is not Python syntax. JSON requires double quotes for its strings.
                try:
                    filters = json.loads(filters_text)
                except json.JSONDecodeError:
                    filters = None
            elif type(filters_text) == dict:
                filters = filters_text
            else:
                filters = None
        return filters  # by default filters are not set

    # 1. QSETS FILTERS (qsets style)

    def _elaborate_qsets_filters(self, data):
        return from_dict_to_qset(data)

    def get_qsets_filters_str(self, request):
        if self.get_qsets_filters(request) is not None:
            return json.dumps(self._decode_text_filters(request, 'filters'))
        else:
            return None

    def get_qsets_filters(self, request):

        filters = self._decode_text_filters(request, 'filters')
        if filters is not None:
            # try:
            return self._elaborate_qsets_filters(filters)
        # except:
        #    print(_('ERROR: qsets structure is incorrect'))
        return None  # by default filters are not set

    # 2. LOCK QSETS FILTERS (qsets style)

    def get_qsets_locked_filters_str(self, request):
        if self.get_qsets_locked_filters(request) is not None:
            return json.dumps(self._decode_text_filters(request, 'locked_filters'))
        else:
            return None

    def get_qsets_locked_filters(self, request):

        filters = self._decode_text_filters(request, 'locked_filters')
        if filters is not None:
            # try:
            return self._elaborate_qsets_filters(filters)
        # except:
        #    print(_('ERROR: qsets structure is incorrect'))
        return None  # by default filters are not set

    # 3. GEO FILTER (dict with keys ['geo_field_name','polygons_srid','polygons'])

    def _elaborate_geo_filter(self, data):
        if issubclass(GEOSGeometry, django.contrib.gis.geos.GEOSGeometry) and \
            issubclass(MultiPolygon, django.contrib.gis.geos.GEOSGeometry):
            if 'geo_field_name' in data and 'polygons_srid' in data and 'polygons' in data:
                geo_field_name = data.get('geo_field_name')
                polygons = []
                for geo_text in data['polygons']:
                    polygons.append(GEOSGeometry(geo_text))
                geo = MultiPolygon(polygons)
                try:
                    geo.srid = int(data.get('polygons_srid'))
                except ValueError:
                    print(_('ERROR: geo srid is incorrect'))
                geo_field = self.model._meta.get_field(geo_field_name)
                _geo = geo.transform(geo_field.srid, clone=True).buffer(0)  # se l'unione del multipolygon risultasse invalida
                qset = Q(**{geo_field_name + '__within': _geo})
            else:
                qset = Q()
            return qset
        return Q()

    def get_geo_filter_str(self, request):
        if self.get_geo_filter(request) is not None:
            return json.dumps(self._decode_text_filters(request, 'geo_filter'))
        else:
            return None

    def get_geo_filter(self, request):
        filters = self._decode_text_filters(request, 'geo_filter')
        if filters is not None:
            try:
                return self._elaborate_geo_filter(filters)
            except:
                print(_('ERROR: geo filter is incorrect'))
        return None  # by default filters are not set

    # 4. SQL FILTERS (list of string)

    def get_sql_filters_str(self, request):
        if self.get_sql_filters(request) is not None:
            return json.dumps(self._decode_text_filters(request, 'sql_filters'))
        else:
            return None

    def get_sql_filters(self, request):
        filters = self._decode_text_filters(request, 'sql_filters')
        if filters is not None:
            return filters
        return None

    # 5. DJANGO WEBIX FILTERS

    def _django_webix_filters_ids(self, request):
        if request.method == 'GET':
            ids = request.GET.getlist('django_webix_filters[]', None)
        elif request.method == 'POST':
            ids = request.POST.getlist('django_webix_filters[]', None)
        else:
            ids = None
        return ids

    def get_django_webix_filters(self, request):
        if self.is_installed_django_webix_filter():
            from django_webix_filter.models import WebixFilter
            filters_ids = self._django_webix_filters_ids(request)
            if filters_ids is not None:
                return WebixFilter.objects.filter(id__in=filters_ids)
        return None

    def get_django_webix_filters_qsets(self):
        if self.is_installed_django_webix_filter():
            qs_django_webix_filter = self.django_webix_filters
            _filter = Q()
            for django_webix_filter in qs_django_webix_filter:
                _filter &= self._elaborate_qsets_filters(django_webix_filter.filter)
            return _filter

        return None

    def _optimize_select_related(self, qs):
        # Estrapolo le informazione per popolare `select_related`
        _select_related = []
        if self.get_fields() is not None:
            for field in self.get_fields():
                if field.get('field_name') is not None:
                    field_name = field.get('field_name')
                    _model = self.model
                    _field = None
                    _related = []
                    for name in field_name.split('__'):
                        try:
                            _field = _model._meta.get_field(name)
                            if isinstance(_field, ManyToManyField):  # Check if field is M2M
                                raise FieldDoesNotExist()
                        except FieldDoesNotExist:
                            break  # name is probably a lookup or transform such as __contains
                        if hasattr(_field, 'related_model') and _field.related_model is not None:
                            _related.append(name)
                            _model = _field.related_model  # field is a relation
                        else:
                            break  # field is not a relation, any name that follows is probably a lookup or transform
                    _related = '__'.join(_related)
                    if _related != '':
                        _select_related.append(_related)
            qs = qs.select_related(*_select_related)
        return qs

    def get_fields(self):
        if self.fields is None:
            return self.fields
        else:
            _fields = []
            for _field in self.fields:
                datalist_column = _field['datalist_column']
                if type(datalist_column)==dict:
                    if 'template_string' in datalist_column:
                        template = Template(datalist_column['template'])
                    elif 'template_name' in datalist_column:
                        template = get_template(datalist_column['template_name'])
                    else:
                        raise Exception('Template is not defined')
                    context = Context(datalist_column.get('context', {}))
                else: # string
                    template = Template(datalist_column)
                    context = Context({})
                _field['datalist_column']  = template.render(context)
                _fields.append(_field)
            return _fields

    def get_queryset(self, initial_queryset=None):
        # bypass improperly configured for custom queryset without model
        if self.model:
            if initial_queryset is None:
                qs = super(WebixListView, self).get_queryset()
            else:
                qs = initial_queryset
            # 1. apply qsets filters
            if self.qsets_filters is not None:
                qs = qs.filter(self.qsets_filters)
            # 2. apply locked qsets filters
            if self.qsets_locked_filters is not None:
                qs = qs.filter(self.qsets_locked_filters)
            # 3. apply geo filter
            if self.geo_filter is not None:
                qs = qs.filter(self.geo_filter)
            # 4. apply SQL raw: this is shit! but need for old SW (remove for future)
            if self.sql_filters is not None:
                sql_filter = ' OR '.join(['({sql})'.format(sql=_sql) for _sql in self.sql_filters])
                qs = qs.extra(where=[sql_filter])
            # 5. apply django webix filters
            #raise Exception(self.django_webix_filters)
            if self.django_webix_filters is not None:
                qs = qs.filter(self.get_django_webix_filters_qsets())

            # optimize select related queryset (only if fields are defined)
            qs = self._optimize_select_related(qs)  # TODO

            return self.apply_ordering(qs)

        return None

    def get_choices_filters(self):
        _fields_choices = {}
        fields = self.get_fields()
        if fields is not None:
            for field in fields:
                if field.get('datalist_column') is not None and \
                    ('serverSelectFilter' in field.get('datalist_column') or
                     'serverRichSelectFilter' in field.get('datalist_column') or
                     'serverMultiSelectFilter' in field.get('datalist_column') or
                     'serverMultiComboFilter' in field.get('datalist_column')):
                    field_name = field.get('field_name')
                    # TODO: there are no null option
                    _fields_choices[field_name] = list(
                        self.get_queryset().filter(**{field_name+'__isnull':False}).values_list(field_name, flat=True).distinct().order_by())

        return _fields_choices

    def is_enable_footer(self):
        is_footer = False
        fields = self.get_fields()
        if fields is not None:
            for field in fields:
                if field.get('footer') is not None:
                    is_footer=True
        return is_footer

    def get_footer(self):
        if self.is_enable_footer():
            qs = self.get_queryset()
            aggregation_dict = {}
            for field in self.get_fields():
                if field.get('footer') is not None:
                    aggregation_dict.update({field.get('field_name') + '_footer': field.get('footer')})
            qs = qs.aggregate(**aggregation_dict)
            return qs
        else:
            return None

    def get_ordering(self):
        if self.request.POST.getlist('sort[]'):
            return self.request.POST.getlist('sort[]')
        else:
            if self.order_by is not None:
                return self.order_by
            else:
                return (self.get_pk_field(),)

    def apply_ordering(self, queryset):
        order = self.get_ordering()
        return queryset.order_by(*order)

    def get_paginate_count(self):
        paginate_count = self.request.GET.get(self.paginate_count_key,
                                              self.request.POST.get(self.paginate_count_key,
                                                                    self.paginate_count_default))
        try:
            return int(paginate_count)
        except ValueError:
            raise Exception(_('Paginate count is not integer'))

    def get_paginate_start(self):
        paginate_start = self.request.GET.get(self.paginate_start_key,
                                              self.request.POST.get(self.paginate_start_key,
                                                                    self.paginate_start_default))
        try:
            return int(paginate_start)
        except ValueError:
            raise Exception(_('Paginate start is not integer'))

    def paginate_queryset(self, queryset):
        paginate_count = self.get_paginate_count()
        paginate_start = self.get_paginate_start()
        return queryset[paginate_start: paginate_start + paginate_count]

    def _get_objects_datatable_values(self, qs):
        values = [self.get_pk_field()]
        fields = self.get_fields()
        if fields is not None:
            for field in fields:
                if field.get('field_name') is not None and \
                    (field.get('queryset_exclude') is None or field.get('queryset_exclude') != True):
                    values.append(field['field_name'])
        data = qs.values(*values,
                         **({'id': F('pk')} if self.get_pk_field() != 'id' and not hasattr(self.model, 'id') else {}))
        return data

    def get_objects_datatable(self):
        if self.model and not self.enable_json_loading:
            qs = self.get_queryset()
            # filters application (like IDS selections)
            qs = self.filters_objects_datatable(qs)
            # pagination (applied only for json request)
            # if self.is_json_request:
            #    qs = self.paginate_queryset(qs)
            # build output
            if qs is not None:
                if type(qs) == list:
                    return qs
                else:  # queryset
                    # apply ordering
                    qs = self.apply_ordering(qs)
                    return self._get_objects_datatable_values(qs)
        return None

    def filters_objects_datatable(self, qs):
        ids = self.request.POST.get('ids', '').split(',')
        ids = list(set(ids) - set([None, '']))

        if len(ids) > 0:
            if type(qs) == list:
                for item in qs:
                    if item.get('id') not in ids:
                        qs.pop(item)
            else:
                auto_field_name = self.model._meta.auto_field.name
                qs = qs.filter(**{auto_field_name + '__in': ids})
        return qs

    def get_pk_field(self):
        if self.pk_field is not None:
            return self.pk_field
        return 'id'

    ########### TEMPLATE BUILDER ###########
    def get_actions(self):
        '''
        Return list ov actions to be executed on listview
        :return:
        '''
        _actions = self.actions

        # add flexport actions
        if apps.is_installed('flexport'):
            from django.contrib.contenttypes.models import ContentType
            from flexport.views import create_extraction
            from flexport.models import Export

            model_ct = ContentType.objects.get(model=self.model._meta.model_name, app_label=self.model._meta.app_label)
            flexport_actions = {}

            def action_builder(export_instance):
                _action = lambda listview, request, qs: create_extraction(request, export_instance.id, qs)
                _action.__name__ = 'flexport_action_%s' % export_instance.id  # verificare se è corretto
                _action.response_type = 'blank'
                _action.short_description = export_instance.action_name
                _action.action_key = 'flexport_{}'.format(export_instance.id)
                _action.allowed_permissions = []
                _action.modal_title = _("Are you sure you want to proceed with this action?")
                _action.modal_ok = _("Proceed")
                _action.modal_cancel = _("Undo")
                return _action

            for export_instance in Export.objects.filter(model=model_ct, active=True):
                if export_instance.is_enabled(self.request):
                    _actions.append(action_builder(export_instance))

        _dict_actions = {}
        for _action in _actions:
            _dict_actions[_action.action_key] = {
                    'func':_action,
                    'action_key': _action.action_key,
                    'response_type': _action.response_type,
                    'allowed_permissions': _action.allowed_permissions,
                    'short_description': _action.short_description,
                    'modal_title': _action.modal_title,
                    'modal_ok': _action.modal_ok,
                    'modal_cancel': _action.modal_cancel,
                }

        return _dict_actions

    def get_actions_style(self):
        _actions_style = None
        if self.actions_style is None:
            _actions_style = 'select'
        elif self.actions_style in ['buttons', 'select']:
            _actions_style = self.actions_style
        else:
            raise ImproperlyConfigured(_(
                "Actions style is improperly configured"
                " only options are 'buttons' or 'select' (select by default)."))
        return _actions_style

    def is_enable_actions(self, request):
        return self.enable_actions

    def is_enable_column_copy(self, request):
        return self.enable_column_copy

    def is_enable_column_delete(self, request):
        return self.enable_column_delete

    def is_enable_row_click(self, request):
        return self.enable_row_click

    def get_type_row_click(self, request):
        return self.type_row_click

    def get_title(self):
        if self.title is not None:
            return self.title
        if self.model is not None:
            return self.model._meta.verbose_name
        return None

    ########### RESPONSE BUILDER ###########

    @property
    def is_json_request(self):
        if 'json' in self.request.GET or 'json' in self.request.POST:
            return True
        else:
            return False

    @property
    def is_action_request(self):
        if 'action' in self.request.GET or 'action' in self.request.POST:
            return True
        else:
            return False

    def get_request_action(self):
        action_name = self.request.POST.get('action', self.request.GET.get('action', None))
        return self.get_actions().get(action_name)['func']


    def post(self, request, *args, **kwargs):  # all post works like get
        return self.get(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        return super(WebixListView, self).get(request, *args, **kwargs)

    def json_response(self, request, *args, **kwargs):
        # ONLY for get_queryset() -> qs and not get_queryset() -> list
        _data = []
        total_count = 0
        if self.model:
            qs = self.get_queryset()
            # filters application (like IDS selections)
            qs = self.filters_objects_datatable(qs)
            # apply ordering
            qs = self.apply_ordering(qs)
            # total count
            total_count = qs.count()
            # apply pagination
            qs_paginate = self.paginate_queryset(qs)
            # build output
            if qs_paginate is not None:
                if type(qs_paginate) == list:
                    raise Exception(
                        _('Json response is available only if get_queryset() return a queryset and not a list'))
                else:  # queryset
                    _data = self._get_objects_datatable_values(qs_paginate)

        pos = self.get_paginate_start()
        # output must be list and not values of queryset
        data = {
            "footer": self.get_footer() if pos==0 else None, # footer is computed only for first page
            'is_enable_footer': self.is_enable_footer(),
            "count": self.get_paginate_count(),
            "total_count": total_count,
            "pos": pos,
            "data": list(_data)
        }
        return JsonResponse(data, safe=False)

    def action_response(self, request, *args, **kwargs):
        action_function = self.get_request_action()
        if action_function is None:
            raise Http404(_('This action is not registered'))
        else:
            # check permissions
            for key in getattr(action_function, 'allowed_permissions', []):
                if hasattr(self, 'has_{}_permission'.format(key)):
                    if getattr(self, 'has_{}_permission'.format(key))(request) != True:
                        raise Http404(_('Permission denied: {}'.format(key)))
                else:
                    raise Http404(_('This permission is not registered on this class'))
            # execution
            qs = self.get_queryset()
            # filters application (like IDS selections)
            qs = self.filters_objects_datatable(qs)
            # apply ordering
            qs = self.apply_ordering(qs)
            return action_function(self, request, qs)

    def dispatch(self, *args, **kwargs):

        # 1. QSETS FILTERS
        self.qsets_filters = self.get_qsets_filters(self.request)

        # 2. QSETS LOCKED FILTERS
        self.qsets_locked_filters = self.get_qsets_locked_filters(self.request)

        # 3. GEO FILTER
        self.geo_filter = self.get_geo_filter(self.request)

        # 4. SQL FILTERS
        self.sql_filters = self.get_sql_filters(self.request)  # for now it's only a qsets list # this is shit! but need for old SW (remove for future)

        # 5. DJANGO WEBIX FILTERS
        self.django_webix_filters = self.get_django_webix_filters(self.request)

        if not self.has_view_permission(request=self.request):
            raise PermissionDenied(_('View permission is not allowed'))

        if self.is_action_request:  # added for action response
            return self.action_response(self.request, *args, **kwargs)
        elif self.is_json_request:  # added for json response with paging
            return self.json_response(self.request, *args, **kwargs)
        else:  # standard response with js webix template structure
            return super(WebixListView, self).dispatch(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(WebixListView, self).get_context_data(**kwargs)
        self.object = None  # bypass for mixin permissions functions
        context.update(self.get_context_data_webix_permissions(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_url(request=self.request, obj=self.object))
        context.update(self.get_context_data_webix_base(request=self.request))
        context.update({
            'fields': self.get_fields(),
            'orders': self.get_ordering(),
            'actions': self.get_actions(),
            'choices_filters': self.get_choices_filters(),
            'footer': self.get_footer() if not self.enable_json_loading else None, # footer only if not paging
            'is_enable_footer': self.is_enable_footer(),
            'get_pk_field': self.get_pk_field(),
            'objects_datatable': self.get_objects_datatable(),
            'is_enable_column_copy': self.is_enable_column_copy(self.request),
            'is_enable_column_delete': self.is_enable_column_delete(self.request),
            'is_enable_row_click': self.is_enable_row_click(self.request),
            'type_row_click': self.get_type_row_click(self.request),
            'is_enable_actions': self.is_enable_actions(self.request),
            'actions_style': self.get_actions_style(),
            'title': self.get_title(),
            # paging
            'is_json_loading': self.enable_json_loading,
            'paginate_count_default': self.paginate_count_default,
            'paginate_count_key': self.paginate_count_key,
            'paginate_start_key': self.paginate_start_key,
            # extra filters
            'is_installed_django_webix_filter': self.is_installed_django_webix_filter(),
            'qsets_filters': self.get_qsets_filters_str(self.request), # for init when template is loaded
            'qsets_locked_filters': self.get_qsets_locked_filters_str(self.request), # for init when template is loaded
            'geo_filter': self.get_geo_filter_str(self.request), # for init when template is loaded
            'sql_filters': self.get_sql_filters_str(self.request), # for init when template is loaded
            'django_webix_filters': self.get_django_webix_filters(self.request), # for init when template is loaded
        })
        return context


class WebixTemplateListView(WebixListView):
    model = None
    enable_column_copy = False
    enable_column_delete = False
    add_permission = False
    change_permission = False
    delete_permission = False
    view_permission = True
    remove_disabled_buttons = True
    enable_row_click = False

    def get_objects_datatable(self):
        raise ImproperlyConfigured(_("Generic TemplateListView needs to define data for datatable"))

    def get_view_prefix(self):
        return ''
